"""Definition of standard version supporting the standard metadata format used by Mojang. 
This module also provide Mojang's version manifest which can be used as default version
repository, allowing resolution of what we call "vanilla" versions.
"""

from subprocess import Popen, TimeoutExpired, PIPE, STDOUT
import xml.etree.ElementTree as ET
from json import JSONDecodeError
from pathlib import Path
from uuid import uuid4
import platform
import shutil
import json
import re
import os

from .download import DownloadList, DownloadEntry, DownloadResultProgress, DownloadResultError
from .util import jvm_bin_filename, merge_dict, calc_input_sha1, LibrarySpecifier
from .auth import AuthSession, OfflineAuthSession
from .http import http_request, HttpError
from . import LAUNCHER_NAME, LAUNCHER_VERSION

from typing import Optional, Iterator, Dict, List, Tuple, Any, Callable, Set


RESOURCES_URL = "https://resources.download.minecraft.net/"
LIBRARIES_URL = "https://libraries.minecraft.net/"
JVM_META_URL = "https://piston-meta.mojang.com/v1/products/java-runtime/2ec0cc96c44e5a76b9c8b7c39df7210883d12871/all.json"
VERSION_MANIFEST_URL = "https://piston-meta.mojang.com/mc/game/version_manifest_v2.json"


class Context:
    """Context of the game's installation and runtime. This defines various directories
    where versions, assets, libraries or JVM are stored, as well as a bin directory for
    temporary runtime files, and also a working directory from where the game will run.
    """

    def __init__(self, 
        main_dir: Optional[Path] = None, 
        work_dir: Optional[Path] = None
    ) -> None:
        """Construct a Minecraft installation context. This context is used by most of the
        installer's tasks to know where to install files and from where to launch the game.

        Note that these paths can perfectly be relative paths, they are computed to 
        absolute paths when needed, so you don't have to care. By default they will be 
        resolved relatively to the current working directory (of the executing Python 
        program).

        :param main_dir: The main directory where versions, assets, libraries and 
        optionally JVM are installed. If not specified this path will be set the usual 
        `.minecraft` (see https://minecraft.wiki/w/.minecraft).
        :param work_dir: The working directory from where the game is run, the game stores
        thing like saves, resource packs, options and mods if relevant. This defaults to
        `main_dir` if not specified.
        """

        main_dir = get_minecraft_dir() if main_dir is None else main_dir
        self.work_dir = main_dir if work_dir is None else work_dir
        self.versions_dir = main_dir / "versions"
        self.assets_dir = main_dir / "assets"
        self.libraries_dir = main_dir / "libraries"
        self.jvm_dir = main_dir / "jvm"
        self.bin_dir = self.work_dir / "bin"

    def get_version(self, version: str) -> "VersionHandle":
        """Get a version's handle.
        """
        return VersionHandle(version, self.versions_dir / version)

    def list_versions(self) -> "Iterator[VersionHandle]":
        """List installed versions given their handles.
        """
        if self.versions_dir.is_dir():
            for version_dir in self.versions_dir.iterdir():
                if version_dir.is_dir():
                    version = VersionHandle(version_dir.name, version_dir)
                    if version.metadata_exists():
                        yield version
    
    def gen_bin_dir(self) -> Path:
        """Generate a random named binary directory, may be used for any kind of temporary
        files and data. Usually for shared libraries used by the game. Note that this 
        directory isn't created by this method, only its path is returned.
        """
        return self.bin_dir / str(uuid4())


class Watcher:
    """Base class for a watcher of the install process.
    """
    
    def handle(self, event: Any) -> None:
        """Called when the watcher can handle the given event. Default implementation
        does nothing.
        """


class VersionHandle:
    """This class holds a version handle that allows modifying its metadata, reading and
    writing it. The parents of this version can also be linked and then merged together
    to get a full metadata.

    This class cannot really be used as a complete version to prepare tasks on, this is
    rather a handle that defines its hierarchical chain of versions.
    """

    __slots__ = "id", "dir", "metadata", "parent"

    def __init__(self, id: str, dir: Path) -> None:
        self.id = id
        self.dir = dir
        self.metadata = {}
        self.parent: Optional[VersionHandle] = None
    
    def metadata_exists(self) -> bool:
        """This function returns true if the version's metadata file exists.
        """
        return self.metadata_file().is_file()
    
    def metadata_file(self) -> Path:
        """This function returns the computed path of the metadata file.
        """
        return self.dir / f"{self.id}.json"

    def jar_file(self) -> Path:
        """This function returns the computed path of the JAR file of the game.
        """
        return self.dir / f"{self.id}.jar"

    def write_metadata_file(self) -> None:
        """This function write the metadata file of the version with the internal data.
        """
        self.dir.mkdir(parents=True, exist_ok=True)
        with self.metadata_file().open("wt") as fp:
            json.dump(self.metadata, fp)

    def read_metadata_file(self) -> bool:
        """This function reads the metadata file and updates the internal data if found.

        :return: True if the data was actually updated from the file.
        """
        try:
            with self.metadata_file().open("rt") as fp:
                self.metadata = json.load(fp)
            return True
        except (OSError, JSONDecodeError):
            return False
    
    def recurse(self) -> "Iterator[VersionHandle]":
        """Walk through every version metadata in the hierarchy of the current one.
        """
        version_meta = self
        while version_meta is not None:
            yield version_meta
            version_meta = version_meta.parent
    
    def merge(self) -> dict:
        """Merge this version metadata and all of its parents into a `FullMetadata`.
        """
        result = {}
        for version_meta in self.recurse():
            merge_dict(result, version_meta.metadata)
        return result
    
    def __str__(self) -> str:
        return self.id
    
    def __repr__(self) -> str:
        return f"<VersionHandle {self.id}>"


class Environment:
    """Describe the game's environment needed to run it. Such instance is produced by
    installing a version and may be used to run the game.

    Note that native libs can contain both JAR files with dynamic libraries inside them 
    and paths to dynamic libraries. The latter being symlinked in the bin directory, or
    copied if not possible (lacking permission on Windows).
    """

    def __init__(self, context: Context, main_class: str) -> None:
        self.context = context
        self.jvm_args: List[str] = []
        self.game_args: List[str] = []
        self.main_class = main_class
        self.args_replacements: Dict[str, str] = {}
        self.native_libs: List[Path] = []
        self.fixes: Dict[str, Any] = {}
    
    def run(self, runner: "Optional[Runner]" = None) -> None:
        """Run this game's environment, with an optional custom runner.
        """
        (runner or StandardRunner()).run(self)

class Runner:
    """Base class handling game running, this default implementation just create a
    process and forwards to its outputs to the outputs of the current process.
    """

    def run(self, env: Environment) -> None:
        raise NotImplementedError


class Version:
    """Base class for basic version resolving, it handles metadata parsing and resources
    resolution. This class provides support for standard versions as provided by the
    Mojang's version manifest.
    """
    
    FIX_LEGACY_PROXY = "legacy_proxy"
    FIX_LEGACY_MERGE_SORT = "legacy_merge_sort"
    FIX_LEGACY_RESOLUTION = "legacy_resolution"
    FIX_LEGACY_QUICK_PLAY = "legacy_quick_play"
    FIX_AUTH_LIB_2_1_28 = "auth_lib_2_1_28"
    FIX_LWJGL = "lwjgl"

    def __init__(self, version: str = "release", *, 
        context: Optional[Context] = None,
    ) -> None:
        """Construct a standard version installer and runner.

        :param context: The installation context of the game, used to know where to find
        its metadata and where to install resources. If the context is not given, the 
        default one is constructed (see Context documentation).
        :param version: The root version to resolve at first, all of its parents will
        be loaded and then all the metadata will be merged together. This merged metadata
        will be used by this class to install its resources such as libraries and assets.
        """

        # Entry attributes
        self.version = version
        self.context = context or Context()
        self.manifest = VersionManifest()
        
        # General options
        self.demo: bool = False
        self.auth_session: Optional[AuthSession] = None
        self.resolution: Optional[Tuple[int, int]] = None
        self.disable_multiplayer: bool = False
        self.disable_chat: bool = False
        self.quick_play: Optional[QuickPlay] = None
        self.jvm_path: Optional[Path] = None
        self.libraries_filters: List[Callable[[Dict[LibrarySpecifier, Library]], None]] = []
        self.fixes: Dict[str, Any] = { 
            self.FIX_LEGACY_PROXY: True, 
            self.FIX_LEGACY_MERGE_SORT: True,
            self.FIX_LEGACY_RESOLUTION: True,
            self.FIX_LEGACY_QUICK_PLAY: True,
            self.FIX_AUTH_LIB_2_1_28: True,
            self.FIX_LWJGL: None
        }

        # Resolved version metadata, root version and its hierarchy, and merged metadata
        self._hierarchy: List[VersionHandle] = []
        self._metadata: dict = {}

        # Features used for rule resolution across the metadata
        self._features: Dict[str, bool] = {}

        # Path to the JAR file to run the game
        self._jar_path: Optional[Path] = None

        # Assets dictionary and index version, with optional legacy directories
        self._assets_index_version: Optional[str] = None
        self._assets: Dict[str, Path] = {}
        self._assets_virtual_dir: Optional[Path] = None
        self._assets_resources_dir: Optional[Path] = None

        # Native and class libraries (using dict so we keep order)
        self._libs: Dict[LibrarySpecifier, Library] = {}
        self._native_libs: List[Path] = []
        self._class_libs: List[Path] = []

        # Logger config parsed from metadata
        self._logger_path: Optional[Path] = None
        self._logger_arg: Optional[str] = None
        
        # JVM to use for launching the game
        self._jvm_path: Optional[Path] = None
        self._jvm_version: Optional[str] = None

        self._dl = DownloadList()
        self._applied_fixes: Dict[str, Any] = {}

    def set_auth_offline(self, username: Optional[str], uuid: Optional[str]) -> None:
        """Shortcut for setting an offline session with the given username/uuid pair.
        """
        self.auth_session = OfflineAuthSession(username, uuid)

    def set_quick_play_singleplayer(self, level_name: str) -> None:
        """Configure quick play for entering a singleplayer level after game's launch.
        """
        self.quick_play = QuickPlaySingleplayer(level_name)
    
    def set_quick_play_multiplayer(self, host: str, port: int = 25565) -> None:
        """Configure quick play for connecting to a server after game's launch.
        """
        self.quick_play = QuickPlayMultiplayer(host, port)
    
    def set_quick_play_realms(self, realm: str) -> None:
        """Configure quick play for connection to a realm after game's launch.
        """
        self.quick_play = QuickPlayRealms(realm)

    def install(self, *, watcher: Optional[Watcher] = None) -> Environment:
        """This function ensures that this version is properly installed given 
        configuration of this class' attributes. This function can be called multiple
        time and will not recompute what have already been done, if errors happen, the
        install can be resumed and the failed steps will be retried.
        """

        watcher = watcher or Watcher()

        self._dl.clear()
        self._applied_fixes.clear()

        self._resolve_version(watcher)
        self._resolve_metadata(watcher)
        self._resolve_features(watcher)
        self._resolve_jvm(watcher)  # JVM added here on purpose to ease implementation of ForgeVersion
        self._resolve_jar(watcher)
        self._resolve_assets(watcher)
        self._resolve_libraries(watcher)
        self._resolve_logger(watcher)
        self._download(watcher)
        self._finalize_assets(watcher)

        return self._resolve_env(watcher)
    
    def _resolve_version(self, watcher: Watcher) -> None:
        """This step is executed just before resolving version metadata and its parents,
        it should process the version and ensure that the version is relevant for the
        metadata resolving.

        The default implementation resolve aliases (release, snapshot) if needed.
        """
        self.version = self.manifest.filter_latest(self.version)[0]

    def _resolve_metadata(self, watcher: Watcher) -> None:
        """This step resolves metadata of the root version and all of its parents.
        """

        hierarchy = self._hierarchy
        version: Optional[str] = self.version
        hierarchy.clear()

        while version is not None:

            if len(hierarchy) > 10:
                raise TooMuchParentsError(list(v.id for v in hierarchy))
            
            watcher.handle(VersionLoadingEvent(version))

            # Get version instance and load/fetch is needed.
            handle = self.context.get_version(version)
            fetched = False
            if not self._load_version(handle, watcher):
                watcher.handle(VersionFetchingEvent(version))
                self._fetch_version(handle, watcher)
                fetched = True
            
            watcher.handle(VersionLoadedEvent(version, fetched))

            # Set the parent of the last version to the version being resolved.
            if len(hierarchy):
                hierarchy[-1].parent = handle
            
            hierarchy.append(handle)
            version = handle.metadata.pop("inheritsFrom", None)
            
            if version is not None and not isinstance(version, str):
                raise ValueError("metadata: /inheritsFrom must be a string")

        self._metadata = hierarchy[0].merge()

    def _load_version(self, version: VersionHandle, watcher: Watcher) -> bool:
        """This function is responsible for loading a version's metadata. Note that 
        implementations are free to load other things beside metadata.

        This function returns true if loading was successful, if this function returns
        false, the `fetch_version` function is then called to fetch the version. This can
        be used to check integrity of a version.

        De default implementation of this function just read metadata file and return 
        true if successful.

        :param version: The version metadata to validate or not.
        :param state: Sequence state when the MetadataTask execute.
        :return: True if the given version is valid and its metadata was properly loaded.
        """

        if not version.read_metadata_file():
            return False

        try:
            version_super_meta = self.manifest.get_version(version.id)
        except HttpError:
            # Silently ignoring HTTP errors, we want to be able to launch offline.
            return True
        
        if version_super_meta is None:
            return True
        else:
            expected_sha1 = version_super_meta.get("sha1")
            if expected_sha1 is not None:
                try:
                    with version.metadata_file().open("rb") as version_meta_fp:
                        current_sha1 = calc_input_sha1(version_meta_fp)
                        return expected_sha1 == current_sha1
                except OSError:
                    return False
            return True

    def _fetch_version(self, version: VersionHandle, watcher: Watcher) -> None:
        """Fetch the data of the given version.

        The default implementation try to fetch version from Mojang's version manifest.

        :param version: The version meta to fetch data into.
        :param state: Sequence state when the MetadataTask execute.
        :raises VersionNotFoundError: In case of error finding the version.
        """

        version_super_meta = self.manifest.get_version(version.id)
        if version_super_meta is None:
            raise VersionNotFoundError(version.id)

        res = http_request("GET", version_super_meta["url"], accept="application/json")
        
        # First decode the data and set it to the version meta. Raising if invalid.
        version.metadata = res.json()
        
        # If successful, write the raw data directly to the file.
        version.dir.mkdir(parents=True, exist_ok=True)
        with version.metadata_file().open("wb") as fp:
            fp.write(res.data)

    def _resolve_features(self, watcher: Watcher) -> None:
        """Step resolving the version's features, whose are a mapping of string to 
        boolean, indicating if such feature is enabled or not. This is later used 
        to compute rules when needed.
        """

        self._features["is_demo_user"] = self.demo
        self._features["has_custom_resolution"] = self.resolution is not None
        if self.quick_play is not None:
            self._features[self.quick_play.feature] = True
        
        watcher.handle(FeaturesEvent([feature for feature, enabled in self._features.items() if enabled]))

    def _resolve_jar(self, watcher: Watcher) -> None:
        """This step resolves the JAR file to use for launcher the game.
        """

        self._jar_path = self._hierarchy[0].jar_file()

        # First try to find a /downloads/client download entry.
        version_dls = self._metadata.get("downloads")
        if version_dls is not None:

            if not isinstance(version_dls, dict):
                raise ValueError("metadata: /downloads must be an object")
            
            client_dl = version_dls.get("client")
            if client_dl is not None:
                self._dl.add(parse_download_entry(client_dl, self._jar_path, "metadata: /downloads/client"), verify=True)
                watcher.handle(JarFoundEvent())
                return
        
        # If no download entry has been found, but the JAR exists, we use it.
        if self._jar_path.is_file():
            watcher.handle(JarFoundEvent())
            return
        
        self._jar_path = None
        raise JarNotFoundError()

    def _resolve_assets(self, watcher: Watcher) -> None:
        """This step resolve assets from metadata and add missing entries to the download
        list for future download.
        """

        metadata = self._metadata
        context = self.context

        assets_index_info = metadata.get("assetIndex")
        if assets_index_info is None:
            # Asset info may not be present, it's not required because some custom 
            # versions may want to use there own internal assets.
            return
        
        if not isinstance(assets_index_info, dict):
            raise ValueError("metadata: /assetIndex must be an object")
        
        assets_index_version = metadata.get("assets", assets_index_info.get("id"))
        if assets_index_version is None:
            # Same as above.
            return
        
        if not isinstance(assets_index_version, str):
            raise ValueError("metadata: /assets or /assetIndex/id must be a string")
        
        watcher.handle(AssetsResolveEvent(assets_index_version, None))

        assets_indexes_dir = context.assets_dir / "indexes"
        assets_index_file = assets_indexes_dir / f"{assets_index_version}.json"

        try:
            with assets_index_file.open("rb") as assets_index_fp:
                assets_index = json.load(assets_index_fp)
        except (OSError, JSONDecodeError):

            # If for some reason we can't read an assets index, try downloading it.
            assets_index_url = assets_index_info.get("url")
            if not isinstance(assets_index_url, str):
                raise ValueError("metadata: /assetIndex/url must be a string")
            
            assets_index = http_request("GET", assets_index_url, accept="application/json").json()
            assets_indexes_dir.mkdir(parents=True, exist_ok=True)
            with assets_index_file.open("wt") as assets_index_fp:
                json.dump(assets_index, assets_index_fp)

        assets_objects_dir = context.assets_dir / "objects"
        assets_resources = assets_index.get("map_to_resources", False)  # For version <= 13w23b
        assets_virtual = assets_index.get("virtual", False)  # For 13w23b < version <= 13w48b (1.7.2)

        if not isinstance(assets_resources, bool):
            raise ValueError("assets index: /map_to_resources must be a boolean")
        if not isinstance(assets_virtual, bool):
            raise ValueError("assets index: /virtual must be a boolean")

        assets_objects = assets_index.get("objects")
        if not isinstance(assets_objects, dict):
            raise ValueError("assets index: /objects must be an object")

        assets_unique = set()

        for asset_id, asset_obj in assets_objects.items():

            if not isinstance(asset_obj, dict):
                raise ValueError(f"assets index: /objects/{asset_id} must be an object")

            asset_hash = asset_obj.get("hash")
            if not isinstance(asset_hash, str):
                raise ValueError(f"assets index: /objects/{asset_id}/hash must be a string")
            
            asset_size = asset_obj.get("size")
            if not isinstance(asset_size, int):
                raise ValueError(f"assets index: /objects/{asset_id}/size must be an integer")

            asset_hash_prefix = asset_hash[:2]
            asset_file = assets_objects_dir.joinpath(asset_hash_prefix, asset_hash)

            self._assets[asset_id] = asset_file

            # Some assets are represented with multiple files, but we don't want to 
            # download a file multiple time so we abort here.
            if asset_hash in assets_unique:
                continue
            assets_unique.add(asset_hash)

            asset_url = f"{RESOURCES_URL}{asset_hash_prefix}/{asset_hash}"
            self._dl.add(DownloadEntry(asset_url, asset_file, size=asset_size, sha1=asset_hash, name=asset_id), verify=True)
        
        self._assets_index_version = assets_index_version
        self._assets_virtual_dir = context.assets_dir.joinpath("virtual", assets_index_version) if assets_virtual else None
        self._assets_resources_dir = context.work_dir / "resources" if assets_resources else None

        watcher.handle(AssetsResolveEvent(assets_index_version, len(self._assets)))

    def _finalize_assets(self, watcher: Watcher) -> None:
        """Step called after download to finalize installation of assets when 
        """

        if self._assets_resources_dir is not None:
            for asset_id, asset_file in self._assets.items():
                dst_file = self._assets_resources_dir / asset_id
                dst_file.parent.mkdir(parents=True, exist_ok=True)
                shutil.copyfile(str(asset_file), str(dst_file))

        if self._assets_virtual_dir is not None:
            for asset_id, asset_file in self._assets.items():
                dst_file = self._assets_virtual_dir / asset_id
                dst_file.parent.mkdir(parents=True, exist_ok=True)
                shutil.copyfile(str(asset_file), str(dst_file))

    def _resolve_libraries(self, watcher: Watcher) -> None:
        """Step resolving libraries from version's metadata. 
        
        **Note that this is the most critical step and libraries resolving is really 
        important for running the game.**
        
        *This step has to support both older format where native libraries were given
        appart from regular class path libraries, all of this should also support 
        automatic downloading both from an explicit artifact URL, or with a maven repo
        URL.*
        """

        watcher.handle(LibrariesResolvingEvent())

        # This set contains all unique library specifiers that have been found so far. 
        # It also contains libs that have been filtered out by rules.
        unique_specs = set()

        # Recursion order is important for libraries resolving, root libraries should
        # be placed first.
        for version in self._hierarchy[0].recurse():

            metadata_libraries = version.metadata.get("libraries")
            if metadata_libraries is None:
                continue

            if not isinstance(metadata_libraries, list):
                raise ValueError("metadata: /libraries must be a list")

            for library_idx, library in enumerate(metadata_libraries):

                if not isinstance(library, dict):
                    raise ValueError(f"metadata: /libraries/{library_idx} must be an object")
                
                name = library.get("name")
                if not isinstance(name, str):
                    raise ValueError(f"metadata: /libraries/{library_idx}/name must be a string")
                
                spec = LibrarySpecifier.from_str(name)

                # Old metadata files provides a 'natives' mapping from OS to the classifier
                # specific for this OS, this kind of libs are "native libs", we need to
                # extract their dynamic libs into the "bin" directory before running.
                natives = library.get("natives")
                if natives is not None:

                    if not isinstance(natives, dict):
                        raise ValueError(f"metadata: /libraries/{library_idx}/natives must be an object")
                    
                    # If natives object is present, the classifier associated to the
                    # OS overrides the lib_spec classifier.
                    spec.classifier = natives.get(minecraft_os)
                    if spec.classifier is None:
                        continue

                    if minecraft_arch_bits is not None:
                        spec.classifier = spec.classifier.replace("${arch}", str(minecraft_arch_bits))

                # Create a wildcard copy of the spec without version, because we don't
                # want to match against version, regardless of its version, a library
                # should not be added twice.
                spec_wild = spec.copy()
                spec_wild.version = "*"

                # We just abort here if the lib name has already been specified. It is
                # really important to do that here after natives resolution because it
                # adds a classifier to the lib spec.
                if spec_wild in unique_specs:
                    continue
                unique_specs.add(spec_wild)

                # Changed in 4.3.1, rules are checked after natives
                rules = library.get("rules")
                if rules is not None:

                    if not isinstance(rules, list):
                        raise ValueError(f"metadata: /libraries/{library_idx}/rules must be a list")
                    
                    if not interpret_rule(rules, self._features, f"metadata: /libraries/{library_idx}/rules"):
                        continue

                lib_entry: Optional[DownloadEntry] = None
                
                # Avoids ready downloading if a fix is being used, in such case we'll use
                # the Mojang's libraries.
                downloads = library.get("downloads")
                if downloads is not None:

                    if not isinstance(downloads, dict):
                        raise ValueError(f"metadata: /libraries/{library_idx}/downloads must be an object")

                    if natives is not None:
                        # Only check classifiers if natives mapping is present.
                        lib_dl_classifiers = downloads.get("classifiers")
                        dl_meta = None if lib_dl_classifiers is None else lib_dl_classifiers.get(spec.classifier)
                    else:
                        # If we are not dealing with natives, just take the artifact.
                        dl_meta = downloads.get("artifact")

                    if dl_meta is not None:
                        # lib_path = self.context.libraries_dir / spec.file_path()
                        lib_entry = parse_download_entry(dl_meta, Path(), f"metadata: /libraries/{library_idx}/downloads/artifact")

                # If no download entry can be found, try to find the maven repository url.
                if lib_entry is None:
                    repo_url = library.get("url")
                    if repo_url is not None:

                        if not isinstance(repo_url, str):
                            raise ValueError(f"metadata: /libraries/{library_idx}/url must be a string")
                        
                        # Let's be sure to have a '/' as last character.
                        if repo_url[-1] != "/":
                            repo_url += "/"
                        
                        lib_path_rel = spec.file_path()
                        # lib_path = self.context.libraries_dir / lib_path_rel
                        lib_entry = DownloadEntry(f"{repo_url}{lib_path_rel}", Path())

                # Adding parsed library.
                # Insertion ordering is guaranteed on dictionaries since python 3.6
                self._libs[spec] = Library(natives is not None, lib_entry)

        # Fix libraries before computation
        self._filter_libraries(watcher)

        # Apply custom libs filters after builtin filters.
        for libs_filter in self.libraries_filters:
            libs_filter(self._libs)

        # Finally take the final version of libs and add them to download list.
        self._native_libs.clear()
        self._class_libs.clear()
        for spec, parsed_lib in self._libs.items():

            lib_path = self.context.libraries_dir / spec.file_path()
            lib_entry = parsed_lib.entry

            # If no repository URL is given, no more download method is available,
            # so if the JAR file isn't installed, the game cannot be launched.
            # 
            # Note: In the past, we used to default the url to Mojang's maven 
            # repository, but this was a bad habit because most libraries could
            # not be downloaded from their repository, and this was confusing to
            # get a download error for such libraries.
            if lib_entry is None or not len(lib_entry.url):
                if not lib_path.is_file():
                    raise LibraryNotFoundError(spec)
            else:
                lib_entry.dst = lib_path
                lib_entry.name = str(spec)
                self._dl.add(lib_entry, verify=True)
            
            # Finally, add it to the correct libs list.
            libs = self._native_libs if parsed_lib.native else self._class_libs
            libs.append(lib_path)

        watcher.handle(LibrariesResolvedEvent(len(self._class_libs), len(self._native_libs)))

    def _filter_libraries(self, watcher: Watcher) -> None:
        """This step should fix libraries and customize the `_libs` list before actually
        registering them for download or check availability. The default implementation is
        used for fixing authlib on 1.16.4 and 1.16.5 and LWJGL fixing.
        """

        # Versions 1.16.4 and 1.16.5 uses authlib:2.1.28 which cause multiplayer button
        # (and probably in-game chat) to be disabled, this can be fixed by switching to
        # version 2.2.30
        if self.fixes.get(self.FIX_AUTH_LIB_2_1_28):
            spec = LibrarySpecifier("com.mojang", "authlib", "2.1.28")
            lib = self._libs.pop(spec, None)
            if lib is not None:
                spec.version = "2.2.30"
                if lib.entry is not None:
                    lib.entry.url = f"{LIBRARIES_URL}com/mojang/authlib/2.2.30/authlib-2.2.30.jar"
                    lib.entry.sha1 = "d6e677199aa6b19c4a9a2e725034149eb3e746f8"
                    lib.entry.size = 87497
                self._applied_fixes[self.FIX_AUTH_LIB_2_1_28] = True
                self._libs[spec] = lib
        
        # Fixing LWJGL, this can be useful on ARM devices because Mojang doesn't provide
        # natives for this architecture.
        lwjgl_version = self.fixes.get(self.FIX_LWJGL)
        if lwjgl_version is not None:

            # Check that we support this LWJGL version.
            if lwjgl_version != "3.2.3" and not lwjgl_version.startswith("3.3."):
                raise ValueError(f"unsupported lwjgl fix version: {lwjgl_version}")

            # Compute natives required for the specific version
            lwjgl_natives_map: Dict[Optional[str], Dict[Optional[str], str]] = {
                "windows": {"x86_64": "natives-windows", "x86": "natives-windows-x86"},
                "linux": {"x86_64": "natives-linux", "x86": "natives-linux", "arm64": "natives-linux-arm64", "arm32": "natives-linux-arm32"},
                "osx": {"x86_64": "natives-macos"}
            }

            # Versions after 3.2.3 supports aarch64 for windows and macOS.
            if lwjgl_version != "3.2.3":
                lwjgl_natives_map["windows"]["arm64"] = "natives-windows-arm64"
                lwjgl_natives_map["osx"]["arm64"] = "natives-macos-arm64"

            # Get natives and error if no natives for current os/arch
            lwjgl_natives = lwjgl_natives_map.get(minecraft_os, {}).get(minecraft_arch)
            if lwjgl_natives is None:
                raise ValueError(f"unsupported lwjgl fix version for your os/arch")

            # Remove all LWJGL libs.
            for to_remove in list(filter(lambda l: l.group == "org.lwjgl", self._libs.keys())):
                del self._libs[to_remove]
            
            def add_lwjgl_lib(name: str) -> None:
                for classifier in (None, lwjgl_natives):
                    spec = LibrarySpecifier("org.lwjgl", name, lwjgl_version, classifier)
                    entry = DownloadEntry(f"https://repo1.maven.org/maven2/{spec.file_path()}", Path())
                    lib = Library(False, entry)
                    self._libs[spec] = lib

            add_lwjgl_lib("lwjgl")
            add_lwjgl_lib("lwjgl-jemalloc")
            add_lwjgl_lib("lwjgl-openal")
            add_lwjgl_lib("lwjgl-opengl")
            add_lwjgl_lib("lwjgl-glfw")
            add_lwjgl_lib("lwjgl-stb")
            add_lwjgl_lib("lwjgl-tinyfd")

            self._applied_fixes[self.FIX_LWJGL] = lwjgl_version

    def _resolve_logger(self, watcher: Watcher) -> None:
        """This step resolve the logger to use for launcher the game.
        """

        logging = self._metadata.get("logging")
        if logging is None:
            return
        
        if not isinstance(logging, dict):
            raise ValueError("metadata: /logging must be an object")
        
        client_logging = logging.get("client")
        if client_logging is None:
            return
        
        if not isinstance(client_logging, dict):
            raise ValueError("metadata: /logging/client must be an object")
        
        argument = client_logging.get("argument")
        if not isinstance(argument, str):
            raise ValueError("metadata: /logging/client/argument must be a string")

        file_info = client_logging.get("file")
        if not isinstance(file_info, dict):
            raise ValueError("metadata: /logging/client/file must be an object")
        
        file_id = file_info.get("id")
        if not isinstance(file_id, str):
            raise ValueError("metadata: /logging/client/file/id must be a string")

        self._logger_arg = argument
        self._logger_path = self.context.assets_dir / "log_configs" / file_id
        dl_entry = parse_download_entry(file_info, self._logger_path, "metadata: /logging/client/file")
        self._dl.add(dl_entry, verify=True)

        watcher.handle(LoggerFoundEvent(file_id.replace(".xml", "")))

    def _resolve_jvm(self, watcher: Watcher) -> None:
        """Step resolving a JVM suitable for running the game.
        """

        watcher.handle(JvmLoadingEvent())

        # Don't do anything if JVM is already provided.
        if self.jvm_path is not None:
            self._jvm_path = self.jvm_path
            self._jvm_version = None
            watcher.handle(JvmLoadedEvent(self._jvm_version, JvmLoadedEvent.CUSTOM))
            return
        
        jvm_version_info = self._metadata.get("javaVersion", {})
        if not isinstance(jvm_version_info, dict):
            raise ValueError("metadata: /javaVersion must be a string")

        jvm_major_version = jvm_version_info.get("majorVersion")
        if jvm_major_version is not None and not isinstance(jvm_major_version, int):
            raise ValueError("metadata: /javaVersion/majorVersion must be an integer")

        if platform.system() == "Linux" and platform.libc_ver()[0] != "glibc":
            return self._resolve_builtin_jvm(watcher, JvmNotFoundError.UNSUPPORTED_LIBC, jvm_major_version)

        jvm_version_type = jvm_version_info.get("component", "jre-legacy")
        if not isinstance(jvm_version_type, str):
            raise ValueError("metadata: /javaVersion/component must be a string")

        jvm_dir = self.context.jvm_dir / jvm_version_type
        jvm_manifest_file = self.context.jvm_dir / f"{jvm_version_type}.json"

        try:
            with jvm_manifest_file.open("rt") as jvm_manifest_fp:
                jvm_manifest = json.load(jvm_manifest_fp)
        except (OSError, JSONDecodeError):

            all_jvm_meta = http_request("GET", JVM_META_URL, accept="application/json").json()
            if not isinstance(all_jvm_meta, dict):
                raise ValueError("jvm metadata: / must be an object")
            
            jvm_arch_meta = all_jvm_meta.get(minecraft_jvm_os)
            if not isinstance(jvm_arch_meta, dict):
                return self._resolve_builtin_jvm(watcher, JvmNotFoundError.UNSUPPORTED_ARCH, jvm_major_version)

            jvm_meta = jvm_arch_meta.get(jvm_version_type)
            if not isinstance(jvm_meta, list) or not len(jvm_meta):
                return self._resolve_builtin_jvm(watcher, JvmNotFoundError.UNSUPPORTED_VERSION, jvm_major_version)

            jvm_meta_manifest = jvm_meta[0].get("manifest")
            if not isinstance(jvm_meta_manifest, dict):
                raise ValueError(f"jvm metadata: /{minecraft_jvm_os}/{jvm_version_type}/0/manifest must be an object")
            
            jvm_meta_manifest_url = jvm_meta_manifest.get("url")
            if not isinstance(jvm_meta_manifest_url, str):
                raise ValueError(f"jvm metadata: /{minecraft_jvm_os}/{jvm_version_type}/0/manifest/url must be a string")

            jvm_manifest = http_request("GET", jvm_meta_manifest_url, accept="application/json").json()

            if not isinstance(jvm_manifest, dict):
                raise ValueError("jvm manifest: / must be an object")

            jvm_manifest["version"] = jvm_meta[0].get("version", {}).get("name")

            jvm_manifest_file.parent.mkdir(parents=True, exist_ok=True)
            with jvm_manifest_file.open("wt") as jvm_manifest_fp:
                json.dump(jvm_manifest, jvm_manifest_fp)
        
        # Special case for macOS because of weird directory structure.
        if minecraft_os == "osx":
            self._jvm_path = jvm_dir.joinpath("jre.bundle/Contents/Home/bin/java")
        else:
            self._jvm_path = jvm_dir.joinpath("bin", jvm_bin_filename)
        
        # This key is custom and set just above in code.
        self._jvm_version = jvm_manifest.get("version")

        jvm_files = jvm_manifest.get("files")
        if not isinstance(jvm_files, dict):
            raise ValueError("jvm manifest: /files must be an object")

        for jvm_file_path_prefix, jvm_file in jvm_files.items():
            if jvm_file.get("type") == "file":

                jvm_file_path = jvm_dir / jvm_file_path_prefix
                jvm_download_raw = jvm_file.get("downloads", {}).get("raw")
                jvm_download_entry = parse_download_entry(jvm_download_raw, jvm_file_path, f"jvm manifest: /files/{jvm_file_path_prefix}/downloads/raw")
                jvm_download_entry.executable = jvm_file.get("executable", False)

                self._dl.add(jvm_download_entry, verify=True)
        
        watcher.handle(JvmLoadedEvent(self._jvm_version, JvmLoadedEvent.MOJANG))

    def _resolve_builtin_jvm(self, watcher: Watcher, reason: str, major_version: Optional[int]) -> None:
        """Internal function to find the builtin Java executable, the reason why this is
        needed is given in parameter. The expected major version is also given, it should
        not be none because we cannot check builtin version.
        """

        if major_version is None:
            raise JvmNotFoundError(reason)

        builtin_path = shutil.which(jvm_bin_filename)
        if builtin_path is None:
            raise JvmNotFoundError(reason)
        
        try:
            
            # Get version of the JVM.
            process = Popen([builtin_path, "-version"], bufsize=1, stdout=PIPE, stderr=STDOUT, universal_newlines=True)
            stdout, _stderr = process.communicate(timeout=1)

            version_start = stdout.index(f"1.{major_version}" if major_version <= 8 else str(major_version))
            version = None
            
            # Parse version by getting all character that are numeric or '.'.
            for i, ch in enumerate(stdout[version_start:]):
                if not ch.isnumeric() and ch not in (".", "_"):
                    version = stdout[version_start:i]
                    break
            
            if version is None:
                raise ValueError()

        except (TimeoutExpired, ValueError):
            raise JvmNotFoundError(JvmNotFoundError.BUILTIN_INVALID_VERSION)

        self._jvm_path = Path(builtin_path)
        self._jvm_version = version
        watcher.handle(JvmLoadedEvent(version, JvmLoadedEvent.BUILTIN))

    def _download(self, watcher: Watcher) -> None:
        
        entries_count = len(self._dl.entries)
        if not entries_count:
            return
        
        # Note: do not create more thread than available entries.
        threads_count = min(entries_count, (os.cpu_count() or 1) * 4)
        errors = []

        watcher.handle(DownloadStartEvent(threads_count, entries_count, self._dl.size))

        for result_count, result in self._dl.download(threads_count, partial_progress=True):
            if isinstance(result, DownloadResultProgress):
                watcher.handle(DownloadProgressEvent(
                    result.thread_id,
                    result_count,
                    result.entry,
                    result.size,
                    result.speed,
                    result.done
                ))
            elif isinstance(result, DownloadResultError):
                errors.append((result.entry, result.code, result.origin))

        # If errors are present, raise an error.
        if len(errors):
            raise DownloadError(errors)
        
        # Clear entries if successful, therefore multiple calls can be chained if
        # needed, without re-downloading the same files.
        self._dl.clear()
        
        watcher.handle(DownloadCompleteEvent())

    def _resolve_env(self, watcher: Watcher) -> Environment:
        """Step for computing correct environment to run the game as configured in this 
        version's instance.
        """

        assert self._assets_index_version is not None, "_resolve_assets() missing"
        assert self._jvm_path is not None, "_resolve_jvm() missing"
        assert self._jar_path is not None, "_resolve_jar() missing"

        # Main class
        main_class = self._metadata.get("mainClass")
        if not isinstance(main_class, str):
            raise ValueError("metadata: /mainClass must be a string")

        # Get authentication of create a random offline.
        auth_session = self.auth_session or OfflineAuthSession(None, None)

        # Class path, without main class (added later depending on arguments present).
        class_path = list(map(lambda path: str(path.absolute()), self._class_libs))

        # Environment definition.
        env = Environment(self.context, main_class)
        env.jvm_args.append(str(self._jvm_path.absolute()))

        env.native_libs = self._native_libs.copy()
        env.fixes = self._applied_fixes
        all_features = set()

        # Check if modern arguments are present (> 1.12.2).
        modern_args = self._metadata.get("arguments")
        if modern_args is not None:

            if not isinstance(modern_args, dict):
                raise ValueError("metadata: /arguments must be an object")
        
            # Interpret JVM arguments.
            modern_jvm_args = modern_args.get("jvm", [])
            interpret_args(modern_jvm_args, self._features, env.jvm_args, "metadata: /arguments/jvm", all_features=all_features)
            
            # Interpret Game arguments.
            modern_game_args = modern_args.get("game", [])
            interpret_args(modern_game_args, self._features, env.game_args, "metadata: /arguments/game", all_features=all_features)
        
        else:

            interpret_args(legacy_jvm_args, self._features, env.jvm_args, "<legacy_jvm_args>", all_features=all_features)

            # Append legacy game arguments, if available.
            legacy_game_args = self._metadata.get("minecraftArguments")
            if legacy_game_args is not None:
                if not isinstance(legacy_game_args, str):
                    raise ValueError("metadata: /minecraftArguments must be a string")
                env.game_args.extend(legacy_game_args.split(" "))

        # JVM argument for logging config
        if self._logger_path is not None and self._logger_arg is not None:
            env.jvm_args.append(self._logger_arg.replace("${path}", str(self._logger_path.absolute())))

        # JVM argument for launch wrapper JAR path
        if main_class == "net.minecraft.launchwrapper.Launch":
            env.jvm_args.append(f"-Dminecraft.client.jar={self._jar_path.absolute()}")

        # If no modern arguments, fix some arguments.
        if modern_args is None:
            # Old versions seems to prefer having the main class first in class path.
            class_path.insert(0, str(self._jar_path.absolute()))
        else:
            # Modern versions seems to prefer having the main class last in class path.
            class_path.append(str(self._jar_path.absolute()))
   
        # Get the last version in the parent's tree, we use it to apply legacy fixes.
        ancestor_id = list(self._hierarchy[0].recurse())[-1].id

        # Legacy proxy aims to fix things like skins on old versions.
        # This is applicable to all alpha/beta and 1.0:1.5
        if self.fixes.get(self.FIX_LEGACY_PROXY):

            proxy_port = None
            if ancestor_id.startswith("a1.0."):
                proxy_port = 80
            elif ancestor_id.startswith("a1.1."):
                proxy_port = 11702
            elif ancestor_id.startswith(("a1.", "b1.")):
                proxy_port = 11705
            elif ancestor_id in ("1.0", "1.1", "1.3", "1.4", "1.5") or \
                ancestor_id.startswith(("1.2.", "1.3.", "1.4.", "1.5.")):
                proxy_port = 11707
            
            if proxy_port is not None:
                self._applied_fixes[self.FIX_LEGACY_PROXY] = f"betacraft.uk:{proxy_port}"
                env.jvm_args.append("-Dhttp.proxyHost=betacraft.uk")
                env.jvm_args.append(f"-Dhttp.proxyPort={proxy_port}")
        
        # Legacy merge sort is applicable to alpha and beta versions.
        if ancestor_id.startswith(("a1.", "b1.")) and self.fixes.get(self.FIX_LEGACY_MERGE_SORT):
            self._applied_fixes[self.FIX_LEGACY_MERGE_SORT] = True
            env.jvm_args.append("-Djava.util.Arrays.useLegacyMergeSort=true")

        # The arguments do not support custom resolution, try to fix.
        if self.resolution is not None and "has_custom_resolution" not in all_features:
            if self.fixes.get(self.FIX_LEGACY_RESOLUTION):
                self._applied_fixes[self.FIX_LEGACY_RESOLUTION] = self.resolution
                env.game_args.extend((
                    "--width", str(self.resolution[0]),
                    "--height", str(self.resolution[1]),
                ))
        
        # The arguments do not support quick play.
        if isinstance(self.quick_play, QuickPlayMultiplayer) and "is_quick_play_multiplayer" not in all_features:
            if self.fixes.get(self.FIX_LEGACY_QUICK_PLAY):
                self._applied_fixes[self.FIX_LEGACY_QUICK_PLAY] = f"{self.quick_play.host}:{self.quick_play.port}"
                env.game_args.extend(("--server", self.quick_play.host))
                env.game_args.extend(("--port", str(self.quick_play.port)))

        # Global options.        
        if self.disable_multiplayer:
            env.game_args.append("--disableMultiplayer")
        if self.disable_chat:
            env.game_args.append("--disableChat")

        # Arguments replacements
        env.args_replacements = {
            # Game
            "auth_player_name": auth_session.username,
            "version_name": self._hierarchy[0].id,
            "library_directory": str(self.context.libraries_dir.absolute()),
            "game_directory": str(self.context.work_dir.absolute()),
            "assets_root": str(self.context.assets_dir.absolute()),
            "assets_index_name": self._assets_index_version,
            "auth_uuid": auth_session.uuid,
            "auth_access_token": auth_session.format_token_argument(False),
            "auth_xuid": auth_session.get_xuid(),
            "clientid": auth_session.client_id,
            "user_type": auth_session.user_type,
            "version_type": self._metadata.get("type", ""),
            # Game (legacy)
            "auth_session": auth_session.format_token_argument(True),
            "game_assets": "" if self._assets_virtual_dir is None else str(self._assets_virtual_dir.absolute()),
            "user_properties": "{}",
            # JVM
            "natives_directory": "",
            "launcher_name": LAUNCHER_NAME,
            "launcher_version": LAUNCHER_VERSION,
            "classpath_separator": os.pathsep,
            "classpath": os.pathsep.join(class_path)
        }

        if self.quick_play is not None and self.quick_play.feature in all_features:
            self.quick_play.add_args_replacements(env.args_replacements)

        if self.resolution is not None:
            env.args_replacements["resolution_width"] = str(self.resolution[0])
            env.args_replacements["resolution_height"] = str(self.resolution[1])
        
        return env


class QuickPlay:
    """Base class for quick play launch methods for the game.
    Note that these quick play types may not be supported by the game.
    """

    feature: str

    def add_args_replacements(self, args_replacements: Dict[str, str]) -> None:
        raise NotImplementedError

class QuickPlaySingleplayer(QuickPlay):
    """Quick play mode to launch a singleplayer level given its name.
    """

    feature = "is_quick_play_singleplayer"

    def __init__(self, level_name: str) -> None:
        self.level_name = level_name

    def add_args_replacements(self, args_replacements: Dict[str, str]) -> None:
        args_replacements["quickPlaySingleplayer"] = self.level_name

class QuickPlayMultiplayer(QuickPlay):
    """Quick play mode to automatically connect to a given server when launching game.
    """

    feature = "is_quick_play_multiplayer"

    def __init__(self, host: str, port: int = 25565) -> None:
        self.host = host
        self.port = port

    def add_args_replacements(self, args_replacements: Dict[str, str]) -> None:
        args_replacements["quickPlayMultiplayer"] = f"{self.host}:{self.port}"

class QuickPlayRealms(QuickPlay):
    """Quick play mode to automatically connection to a given realm when launching game.
    """

    feature = "is_quick_play_realms"

    def __init__(self, realm: str) -> None:
        self.realm = realm
    
    def add_args_replacements(self, args_replacements: Dict[str, str]) -> None:
        args_replacements["quickPlayRealms"] = self.realm


class WatcherGroup(Watcher):
    """A group of watcher that is itself a watcher, its functions dispatches events to
    all tasks.
    """

    def __init__(self) -> None:
        self.children: Set[Watcher] = set()
    
    def add(self, watcher: Watcher) -> None:
        """Add a watcher to the installer to this group.
        """
        self.children.add(watcher)
    
    def remove(self, watcher: Watcher) -> None:
        """Remove a watcher from the group.
        """
        self.children.remove(watcher)
    
    def handle(self, event: Any) -> None:
        for watcher in self.children:
            watcher.handle(event)

class SimpleWatcher(Watcher):

    def __init__(self, handlers: Dict[type, Callable[[Any], None]]) -> None:
        self.handlers = handlers

    def handle(self, event: Any) -> None:
        handler = self.handlers.get(type(event))
        if handler is not None:
            handler(event)


class Library:
    """Intermediate class representing a parsed game's library.
    """
    __slots__ = "native", "entry"
    def __init__(self, native: bool, entry: Optional[DownloadEntry]) -> None:
        self.native = native
        self.entry = entry


class VersionNotFoundError(Exception):
    """Raised when a version was not found. The version that was not found is given.
    """
    def __init__(self, version: str) -> None:
        self.version = version
    
    def __str__(self) -> str:
        return repr(self.version)

class TooMuchParentsError(Exception):
    """Raised when a version hierarchy is too deep. The hierarchy of versions is given
    in property `versions`.
    """
    def __init__(self, versions: List[str]) -> None:
        self.versions = versions

    def __str__(self) -> str:
        return repr(self.versions)

class LibraryNotFoundError(Exception):
    """Critical error raised when a library has no download indication and is not 
    currently installed in game's libraries.
    """
    def __init__(self, lib: LibrarySpecifier) -> None:
        self.lib = lib
    
    def __str__(self) -> str:
        return repr(self.lib)

class JarNotFoundError(Exception):
    """Raised when no version's JAR file could be found from the metadata.
    """

class JvmNotFoundError(Exception):
    """Raised if no JVM can be found, the particular reason is given as code. This error
    is raised only if no builtin Java can be resolved.
    """

    UNSUPPORTED_LIBC = "unsupported_libc"
    UNSUPPORTED_ARCH = "unsupported_arch"
    UNSUPPORTED_VERSION = "unsupported_version"
    BUILTIN_INVALID_VERSION = "builtin_invalid_version"

    def __init__(self, code: str) -> None:
        self.code = code
    
    def __str__(self) -> str:
        return repr(self.code)

class DownloadError(Exception):
    """Raised when the downloader failed to download some entries.
    """
    def __init__(self, errors: List[Tuple[DownloadEntry, str, Optional[Exception]]]) -> None:
        self.errors = errors
    
    def __str__(self) -> str:
        return repr(self.errors)


class VersionEvent:
    """Base class for events regarding version.
    """
    __slots__ = "version",
    def __init__(self, version: str) -> None:
        self.version = version

class VersionLoadingEvent(VersionEvent):
    """Event triggered when a version is being loaded.
    """
    __slots__ = tuple()

class VersionFetchingEvent(VersionEvent):
    """Event triggered when a version is being fetched.
    """
    __slots__ = tuple()

class VersionLoadedEvent(VersionEvent):
    """Event triggered when a version has been successfully loaded.
    """
    __slots__ = "fetched",
    def __init__(self, version: str, fetched: bool) -> None:
        super().__init__(version)
        self.fetched = fetched

class FeaturesEvent:
    """Event triggered when features for the version has been computed. Only enabled 
    features are given as list of features.
    """
    __slots__ = "features",
    def __init__(self, features: List[str]) -> None:
        self.features = features

class JarFoundEvent:
    """Event triggered when the game's JAR file has been found.
    """
    __slots__ = tuple()

class AssetsResolveEvent:
    __slots__ = "index_version", "count"
    def __init__(self, index_version: str, count: Optional[int]) -> None:
        self.index_version = index_version
        self.count = count

class LibrariesResolvingEvent:
    """Event triggered when libraries start being resolved.
    """
    __slots__ = tuple()

class LibrariesResolvedEvent:
    """Event triggered when all libraries has been successfully resolved.
    """
    __slots__ = "class_libs_count", "native_libs_count"
    def __init__(self, class_libs_count: int, native_libs_count: int) -> None:
        self.class_libs_count = class_libs_count
        self.native_libs_count = native_libs_count

class LoggerFoundEvent:
    __slots__ = "version"
    def __init__(self, version: str) -> None:
        self.version = version

class JvmLoadingEvent:
    """Event triggered when JVM start being resolved.
    """
    __slots__ = tuple()

class JvmLoadedEvent:
    """Event triggered when JVM has been resolved. If count is none then the resolved 
    version is a builtin JVM.
    """

    MOJANG = "mojang"    # Mojang provided JVM
    BUILTIN = "builtin"  # Builtin JVM (java command)
    CUSTOM = "custom"    # Custom JVM given with jvm_path

    __slots__ = "version", "kind"
    def __init__(self, version: Optional[str], kind: str) -> None:
        self.version = version
        self.kind = kind

class DownloadStartEvent:
    __slots__ = "threads_count", "entries_count", "size"
    def __init__(self, threads_count: int, entries_count: int, size: int) -> None:
        self.threads_count = threads_count
        self.entries_count = entries_count
        self.size = size

class DownloadProgressEvent:
    __slots__ = "thread_id", "count", "entry", "size", "speed", "done"
    def __init__(self, thread_id: int, count: int, entry: DownloadEntry, size: int, speed: float, done: bool) -> None:
        self.thread_id = thread_id
        self.count = count
        self.entry = entry
        self.size = size
        self.speed = speed
        self.done = done

class DownloadCompleteEvent:
    __slots__ = tuple()


class VersionManifest:
    """The Mojang's official version manifest. Providing officially available versions 
    with optional cache file.
    """

    def __init__(self, cache_file: Optional[Path] = None) -> None:
        self.data: Optional[dict] = None
        self.cache_file = cache_file

    def _ensure_data(self) -> dict:
        """Internal method that ensure that the manifest data is up-to-date.

        :return: The full data of the manifest.
        :raises HttpError: Underlying HTTP error if manifest could not be requested.
        """

        if self.data is None:

            headers = {}
            cache_data = None

            # If a cache file should be used, try opening it and read the last modified
            # time that will be used for requesting the manifest, only if needed.
            if self.cache_file is not None:
                try:
                    with self.cache_file.open("rt") as cache_fp:
                        cache_data = json.load(cache_fp)
                    if isinstance(cache_data, dict):
                        if "last_modified" in cache_data:
                            headers["If-Modified-Since"] = cache_data["last_modified"]
                    else:
                        # If the data isn't a dictionary, it's invalid.
                        cache_data = None
                except (OSError, json.JSONDecodeError):
                    pass
            
            try:

                res = http_request("GET", VERSION_MANIFEST_URL, 
                    headers=headers, 
                    accept="application/json")
                
                self.data = res.json()
                if not isinstance(self.data, dict):
                    # Raise error because the API is faulting.
                    raise ValueError("manifest api data is not dict")

                if "Last-Modified" in res.headers:
                    self.data["last_modified"] = res.headers["Last-Modified"]

                if self.cache_file is not None:
                    self.cache_file.parent.mkdir(parents=True, exist_ok=True)
                    with self.cache_file.open("wt") as cache_fp:
                        json.dump(self.data, cache_fp)

            except HttpError as error:
                # Checking for 0, which means network error, in such case we want to 
                # ignore the network error and just use the cached data.
                if error.res.status in (0, 304) and cache_data is not None:
                    self.data = cache_data
                else:
                    raise

        if self.data is None:
            raise ValueError("manifest data should not be null at this point")
        
        return self.data

    def is_alias(self, version: str) -> bool:
        """Basic function that returns true if the given version is an release or
        snapshot alias.
        """
        return version in ("release", "snapshot")

    def filter_latest(self, version: str) -> Tuple[str, bool]:
        """Filter a version identifier if 'release' or 'snapshot' alias is used, then it's
        replaced by the full version identifier, like `1.19.3`.

        :param version: The version id or alias.
        :return: A tuple containing the full version id and a boolean indicating if the
        given version identifier is an alias.
        :raises HttpError: Underlying HTTP error if manifest could not be requested, only
        possible when the given version is a known alias.
        """

        if self.is_alias(version):
            latest = self._ensure_data()["latest"].get(version)
            if latest is not None:
                return latest, True
        return version, False

    def get_version(self, version: str) -> Optional[dict]:
        """Get a manifest's version metadata. Containing the metadata's URL, its SHA1 and
        its type.

        :param version: The version identifier.
        :return: If found, the version is returned.
        :raises HttpError: Underlying HTTP error if manifest could not be requested.
        """
        version, _alias = self.filter_latest(version)
        for version_data in self._ensure_data()["versions"]:
            if version_data["id"] == version:
                return version_data
        return None

    def all_versions(self) -> list:
        return self._ensure_data()["versions"]


class StandardRunner(Runner):
    """Base class handling game running, this default implementation just create a
    process and forwards to its outputs to the outputs of the current process. This 
    runner supports KeyboardInterrupt handling.
    """

    def run(self, env: Environment) -> None:

        from zipfile import ZipFile

        bin_dir = env.context.gen_bin_dir().absolute()
        replacements = env.args_replacements.copy()
        replacements["natives_directory"] = str(bin_dir)
        
        bin_dir.mkdir(parents=True, exist_ok=True)

        try:

            # Here we copy libraries into the bin directory, in case of archives (jar, zip)
            # we extract all so/dll/dylib files into the directory, if this is a directly
            # pointing to an archive, we symlink or copy it in-place.
            if len(env.native_libs):
                for src_file in env.native_libs:

                    if not src_file.is_file():
                        raise ValueError(f"source native file not found: {src_file}")

                    native_name = src_file.name
                    if native_name.endswith((".zip", ".jar")):

                        with ZipFile(src_file, "r") as native_zip:
                            for native_zip_info in native_zip.infolist():
                                native_name = native_zip_info.filename
                                if native_name.endswith((".so", ".dll", ".dylib")):

                                    try:
                                        native_name = native_name[native_name.rindex("/") + 1:]
                                    except ValueError:
                                        native_name = native_name
                                    
                                    dst_file = bin_dir / native_name

                                    with native_zip.open(native_zip_info, "r") as src_fp:
                                        with dst_file.open("wb") as dst_fp:
                                            shutil.copyfileobj(src_fp, dst_fp)
                                    
                    else:

                        # Here we try to remove the version numbers of .so files.
                        so_idx = native_name.rfind(".so")
                        if so_idx >= 0:
                            native_name = native_name[:so_idx + len(".so")]
                        # Try to symlink the file in the bin dir, and fallback to simple copy.
                        dst_file = bin_dir / native_name

                        try:
                            dst_file.symlink_to(src_file)
                        except OSError:
                            shutil.copyfile(src_file, dst_file)
                        
            # We create the wrapper process with required arguments.
            process = self.process_create([
                *replace_list_vars(env.jvm_args, replacements),
                env.main_class,
                *replace_list_vars(env.game_args, replacements)
            ], env.context.work_dir)

            if process is not None:
                self.process_wait(process)

        finally:
            # Any error while setting up the binary directory cause it to be deleted.
            shutil.rmtree(bin_dir, ignore_errors=True)

    def process_create(self, args: List[str], work_dir: Path) -> Optional[Popen]:
        """This function is called when process needs to be created with the given 
        arguments in the given working directory. The default implementation does nothing
        special but this can be used to create the process with enabled output piping,
        to later use in `process_wait`.

        None can be returned to abort starting the game.
        """
        return Popen(args, cwd=work_dir)

    def process_wait(self, process: Popen) -> None:
        """This function is called with the running Minecraft process for waiting the end
        of the process. Implementors may want to read incoming logging.
        """
        import time
        try:
            while process.poll() is None:
                time.sleep(1)
        except KeyboardInterrupt:
            process.kill()
            raise
        finally:
            process.wait()


class StreamRunner(StandardRunner):
    """A specialized implementation of `StandardRunner` which allows streaming the game's 
    output logs. This implementation also provides parsing of log4j XML layouts for logs. 
    This runner handles KeyboardInterrupt errors and properly kill the game and waits for 
    its completion.
    """
    
    def process_create(self, args: List[str], work_dir: Path) -> Optional[Popen]:
        return Popen(args, cwd=work_dir, stdout=PIPE, stderr=STDOUT, bufsize=1, universal_newlines=True, encoding="utf-8", errors="replace")

    def process_wait(self, process: Popen) -> None:

        from threading import Thread
        import time

        thread = Thread(target=self.process_stream_thread, name="Minecraft Stream Thread", args=(process,), daemon=True)
        thread.start()

        try:
            while thread.is_alive():
                time.sleep(1)
        except KeyboardInterrupt:
            process.kill()
            raise
        finally:
            process.wait()

    def process_stream_thread(self, process: Popen) -> None:

        stdout = process.stdout
        assert stdout is not None, "should not be none because it should be piped"

        parser = StreamParser()
        for line in iter(stdout.readline, ""):

            # By default the parser is raw but if we encounter any potential XML tag we
            # set the parser to the XML one and starting parsing. If it fails immediately
            # then it reset it to a raw stream parser.
            if line.lstrip().startswith("<log4j:") and not isinstance(parser, XmlStreamParser):
                parser = XmlStreamParser()

            if not parser.feed(line, self.process_stream_event):
                parser = StreamParser()
                parser.feed(line, self.process_stream_event)  # Should not fail
    
    def process_stream_event(self, event: Any) -> None:
        """This function gets called when an event is received from the game's log.
        """

class StreamParser:
    """Base implementation of game's output stream parsing.
    """

    def feed(self, line: str, callback: Callable[[Any], None]) -> bool:
        """Feed the parser, if successful the callback can be used to give back the object
        containing the parsed log, and true should be returned. Returning false mean that
        parsing has failed, the caller can then switch to a simpler parser.

        This default implementation just forward incoming lines to the callback.
        """
        callback(line)
        return True

class XmlStreamParser(StreamParser):
    """This parser produces `XmlStreamEvent` kind of events by parsing the game's stream
    as a log4j log stream.
    """

    def __init__(self) -> None:
        self.xml = ET.XMLPullParser(["start", "end"])
        self.xml.feed("<?xml version=\"1.0\"?><root xmlns:log4j=\"log4j\">")
        self.next_event = None

    def feed(self, line: str, callback: Callable[[Any], None]) -> bool:
        self.xml.feed(line)
        try:
            for event, elem in self.xml.read_events():
                if elem.tag == "{log4j}Event":
                    if event == "start":
                        self.next_event = XmlStreamEvent(int(elem.attrib["timestamp"]) / 1000.0,
                            elem.attrib["logger"],
                            elem.attrib["level"],
                            elem.attrib["thread"])
                    elif event == "end" and self.next_event is not None:
                        callback(self.next_event)
                        self.next_event = None
                elif event == "end" and self.next_event is not None:
                    if elem.tag == "{log4j}Message":
                        self.next_event.message = elem.text
                    elif elem.tag == "{log4j}Throwable":
                        self.next_event.throwable = elem.text
            return True
        except ET.ParseError:
            return False

class XmlStreamEvent:
    """Class representing an event happening in the game's logs.
    """

    __slots__ = "time", "logger", "level", "thread", "message", "throwable"

    def __init__(self, time: float, logger: str, level: str, thread: str) -> None:
        self.time = time
        self.logger = logger
        self.level = level
        self.thread = thread
        self.message = None
        self.throwable = None
    
    def __repr__(self) -> str:
        return f"<ProcessEvent date: {self.time}, logger: {self.logger}, level: {self.level}, thread: {self.thread}, message: {repr(self.message)}>"


def parse_download_entry(value: Any, dst: Path, path: str) -> DownloadEntry:
    """Common function to parse a download entry from a metadata JSON file.
    """

    if not isinstance(value, dict):
        raise ValueError(f"{path} must an object")

    url = value.get("url")
    if not isinstance(url, str):
        raise ValueError(f"{path}/url must be a string")

    size = value.get("size")
    if size is not None and not isinstance(size, int):
        raise ValueError(f"{path}/size must be an integer")

    sha1 = value.get("sha1")
    if sha1 is not None and not isinstance(sha1, str):
        raise ValueError(f"{path}/sha1 must be a string")

    return DownloadEntry(url, dst, size=size, sha1=sha1, name=dst.name)


def interpret_rule(rules: Any, features: Dict[str, bool], path: str, *, 
    all_features: Optional[Set[str]] = None
) -> bool:
    """Common function to interpret rules and determine if the condition is met.
    """

    if not isinstance(rules, list):
        raise ValueError(f"{path} must be a list")

    allowed = False
    for i, rule in enumerate(rules):

        if not isinstance(rule, dict):
            raise ValueError(f"{path}/{i} must be an object")

        rule_os = rule.get("os")
        if rule_os is not None and not interpret_rule_os(rule_os, f"{path}/{i}/os"):
            continue

        rule_features = rule.get("features")
        if rule_features is not None:
            
            if not isinstance(rule_features, dict):
                raise ValueError(f"{path}/{i}/features must be an object")

            feat_valid = True
            for feat_name, feat_expected in rule_features.items():
                if all_features is not None:
                    all_features.add(feat_name)
                if features.get(feat_name) != feat_expected:
                    feat_valid = False
            
            if not feat_valid:
                continue
        
        action = rule.get("action")
        if action == "disallow":
            return False    # Early return because of disallow.
        elif action == "allow":
            allowed = True  # Only other possible value is "allow".
        else:
            raise ValueError(f"{path}/{i}/action must be 'allow' and 'disallow'")

    return allowed


def interpret_rule_os(rule_os: Any, path: str) -> bool:
    """Common function to interpret a rule constraint on the running OS.
    """

    if not isinstance(rule_os, dict):
        raise ValueError(f"{path} must be an object")
    
    os_name = rule_os.get("name")
    if os_name is None or os_name == minecraft_os:
        os_arch = rule_os.get("arch")
        if os_arch is None or os_arch == minecraft_arch:
            os_version = rule_os.get("version")
            if os_version is None or re.search(os_version, platform.version()) is not None:
                return True
    return False


def interpret_args(args: Any, features: Dict[str, bool], dst: List[str], path: str, *, 
    all_features: Optional[Set[str]] = None
) -> None:
    """Common function for interpreting a list of arguments, whose may be conditional
    under some rules. An optional set of features can be given and will be filled with
    all features found (even if not used).
    """

    if not isinstance(args, list):
        raise ValueError(f"{path} must be a list")

    for i, arg in enumerate(args):

        if isinstance(arg, str):
            dst.append(arg)
        elif isinstance(arg, dict):

            rules = arg.get("rules")
            if rules is not None:
                if not interpret_rule(rules, features, f"{path}/{i}/rules", all_features=all_features):
                    continue
            
            arg_value = arg["value"]
            if isinstance(arg_value, list):
                dst.extend(arg_value)
            elif isinstance(arg_value, str):
                dst.append(arg_value)
            else:
                raise ValueError(f"{path}/{i}/value must be a list or a string")
        else:
            raise ValueError(f"{path}/{i} must be an object or a string")


def replace_vars(text: str, replacements: Dict[str, str]) -> str:
    """Replace all variables of the form `${foo}` in a string. If some keys are missing,
    the unformatted text is returned.
    """
    try:
        return text.replace("${", "{").format_map(replacements)
    except KeyError:
        return text


def replace_list_vars(text_list: List[str], replacements: Dict[str, str]) -> Iterator[str]:
    """Call `replace_vars` on multiple texts in a list with the same replacements.
    """
    return (replace_vars(elt, replacements) for elt in text_list)


def get_minecraft_dir() -> Path:
    """Internal function to get the default directory for installing
    and running Minecraft.
    """
    home = Path.home()
    return {
        "Windows": home.joinpath("AppData", "Roaming", ".minecraft"),
        "Darwin": home.joinpath("Library", "Application Support", "minecraft"),
    }.get(platform.system(), home / ".minecraft")


# Name of the OS has used by Minecraft.
minecraft_os = {
    "Linux": "linux", 
    "Windows": "windows", 
    "Darwin": "osx",
    "FreeBSD": "freebsd"
}.get(platform.system())

# Name of the processor's architecture has used by Minecraft.
minecraft_arch = {
    "i386": "x86",
    "i686": "x86",
    "x86_64": "x86_64",
    "amd64": "x86_64",
    "arm64": "arm64",
    "aarch64": "arm64",
    "armv7l": "arm32",
    "armv6l": "arm32",
}.get(platform.machine().lower())

# Stores the bits length of pointers on the current system.
minecraft_arch_bits = {
    "64bit": 64,
    "32bit": 32
}.get(platform.architecture()[0])

# Name of the OS has used by Mojang for officially distributed JVMs.
minecraft_jvm_os = None if minecraft_arch is None else {
    "Darwin": {"x86_64": "mac-os", "arm64": "mac-os-arm64"},
    "Linux": {"x86": "linux-i386", "x86_64": "linux"},
    "Windows": {"x86": "windows-x86", "x86_64": "windows-x64"}
}.get(platform.system(), {}).get(minecraft_arch)

# JVM arguments used if no arguments are specified.
legacy_jvm_args = [
    {
        "rules": [{"action": "allow", "os": {"name": "osx"}}],
        "value": ["-XstartOnFirstThread"]
    },
    {
        "rules": [{"action": "allow", "os": {"name": "windows"}}],
        "value": "-XX:HeapDumpPath=MojangTricksIntelDriversForPerformance_javaw.exe_minecraft.exe.heapdump"
    },
    {
        "rules": [{"action": "allow", "os": {"name": "windows", "version": "^10\\."}}],
        "value": ["-Dos.name=Windows 10", "-Dos.version=10.0"]
    },
    "-Djava.library.path=${natives_directory}",
    "-Dminecraft.launcher.brand=${launcher_name}",
    "-Dminecraft.launcher.version=${launcher_version}",
    "-cp",
    "${classpath}"
]



"""Utilities specific to formatting the output of the CLI.
"""

from .lang import get_raw as _raw

import shutil
import time
import sys
import re

from typing import List, Tuple, Union, Optional


class OutputTable:
    """Base class for formatting tables.
    """

    def __init__(self) -> None:
        self.rows: List[Union[None, Tuple[str, ...]]] = []
        self.columns_length: List[int] = []

    def add(self, *cells):
        """Add a row to the table.
        """

        cells_str = tuple(map(str, cells))
        self.rows.append(cells_str)

        columns_count = len(self.columns_length)
        
        # Limit to the current columns count.
        for i, cell in enumerate(cells_str[:columns_count]):
            cell_len = len(cell)
            if self.columns_length[i] < cell_len:
                self.columns_length[i] = cell_len

        # Adjust if the inserted cells have more columns that current state. 
        if columns_count < len(cells_str):
            self.columns_length.extend(map(len, cells_str[columns_count:]))
    
    def separator(self) -> None:
        """Add a separator to the table.
        """
        self.rows.append(None)
    
    def print(self) -> None:
        """Print the table to the output.
        """
        raise NotImplementedError


class Output:
    """This class is used to abstract the output of the CLI. This particular class is
    abstract and the implementation differs depending on the desired output format.
    """

    def table(self) -> OutputTable:
        """Create a table builder that you can use to add rows and separator and them
        print a table, adapted to the implementation.
        """
        raise NotImplementedError
    
    def task(self, state: Optional[str], key: Optional[str], **kwargs) -> None:
        """Update the current task (or create it if not the case).
        """
        raise NotImplementedError
    
    def finish(self) -> None:
        """Finish any active task.
        """
        raise NotImplementedError

    def print(self, text: str) -> None:
        """Raw print of the given text, this is commonly used to forward game's standard
        output/error streams. The implementor may apply some style before printing text.
        This function doesn't add any new line
        """
        raise NotImplementedError

    def prompt(self, password: bool = False) -> Optional[str]:
        """Prompt for a line to come on standard input.
        """
        raise NotImplementedError


class HumanOutput(Output):

    state_colors = {
        "OK": "\033[92m",
        "FAILED": "\033[31m",
        "WARN": "\033[33m",
        "INFO": "\033[34m",
        "HALT": "\033[33m",
    }

    print_colors = [
        ("ERROR", "\033[31m"),
        ("WARN", "\033[33m"),
        ("SEVERE", "\033[31m"),
        ("FATAL", "\033[31m"),
    ]

    def __init__(self, color: bool) -> None:
        super().__init__()
        self.term_width = 0
        self.term_width_update_time = 0
        self.last_len = None
        self.color = color

    def get_term_width(self) -> int:
        """Internal method used to get terminal width with a cache interval of 1 second.
        """
        now = time.monotonic()
        if now - self.term_width_update_time > 1:
            self.term_width_update_time = now
            self.term_width = shutil.get_terminal_size().columns
        return self.term_width

    def table(self) -> OutputTable:
        return HumanTable(self)
    
    def task(self, state: Optional[str], key: Optional[str], **kwargs) -> None:

        # Don't display updates on small terminals (9 for the state, 11 for msg).
        term_width = self.get_term_width()
        if term_width < 20:
            return
        
        # Get header for the given state (with optional color).
        if state is None:
            state_msg = "\r         "
        else:
            color = self.state_colors.get(state) if self.color else None
            if color is not None:
                state_msg = f"\r[{color}{state:^6s}\033[0m] "
            else:
                state_msg = f"\r[{state:^6s}] "

        print(state_msg, end="", flush=False)

        if key is None:
            self.last_len = 0
            sys.stdout.flush()
            return

        msg = _raw(key, kwargs)
        if len(msg) + 9 > term_width:
            msg = f"{msg[:term_width - 9 - 3]}..."
        
        msg_len = len(msg)

        print(msg, end="", flush=False)

        if self.last_len is not None and self.last_len > msg_len:
            missing_len = self.last_len - msg_len
            print(" " * missing_len, end="", flush=False)
        
        sys.stdout.flush()

        self.last_len = msg_len

    def finish(self) -> None:
        if self.last_len is not None:
            print()
            self.last_len = None
    
    def print(self, text: str) -> None:
        
        if self.color:

            chosen_color = None
            for token, code in self.print_colors:
                if token in text:
                    chosen_color = code
                    break
            
            if chosen_color is not None:
                print(chosen_color, text, "\033[0m", sep="", end="", flush=True)
                return
        
        print(text, end="", flush=True)
    
    def prompt(self, password: bool = False) -> Optional[str]:
        try:
            if password:
                import getpass
                return getpass.getpass("")
            else:
                return input("")
        except KeyboardInterrupt:
            return None

class HumanTable(OutputTable):
    
    def __init__(self, out: HumanOutput) -> None:
        super().__init__()
        self.out = out

    def print(self) -> None:

        columns_length = self.columns_length.copy()
        columns_count = len(columns_length)

        total_length = 1 + sum(x + 3 for x in columns_length)
        max_length = self.out.get_term_width() - 1
        if total_length > max_length:
            
            overflow_length = total_length - max_length
            total_columns_length = sum(columns_length)

            for i in range(columns_count):
                column_ratio = columns_length[i] / max_length
                cell_overflow_length = int(columns_length[i] / total_columns_length * column_ratio * overflow_length)
                overflow_length -= cell_overflow_length
                columns_length[i] -= cell_overflow_length
                if i == columns_count - 1:
                    columns_length[i] -= overflow_length
        
        format_string = "│ {} │".format(" │ ".join((f"{{:{length}s}}" for length in columns_length)))
        columns_lines = ["─" * length for length in columns_length]

        print("┌─{}─┐".format("─┬─".join(columns_lines)), flush=False)

        format_columns = [""] * columns_count

        for row in self.rows:

            if row is None:
                print("├─{}─┤".format("─┼─".join(columns_lines)), flush=False)
                continue

            wrapped_row = list(row)
            wrapped = True

            while wrapped:
                wrapped = False
                for col_index, col in enumerate(wrapped_row):
                    col_len = columns_length[col_index]
                    col_real = col[:col_len]
                    format_columns[col_index] = col_real
                    # If wrapped, take the rest and save it for next iteration.
                    if col != col_real:
                        wrapped_row[col_index] = f" {col[col_len:]}"
                        wrapped = True
                    else:
                        wrapped_row[col_index] = ""
                
                print(format_string.format(*format_columns), flush=False)
        
        print("└─{}─┘".format("─┴─".join(columns_lines)), flush=True)


class MachineOutput(Output):

    escape_re = re.compile("[\\n\\r,]")

    @classmethod
    def print_escape(cls, s: str) -> str:
        return re.sub(cls.escape_re, lambda match: "\\" + {10: "n", 13: "r"}.get(ord(match.group()), match.group()), s)

    def print_function(self, name: str, *args: str, **kwargs) -> None:
        """Print a machine-readable line for a function with some parameters.
        """
        print(name, ":", ",".join((self.print_escape(arg) for arg in [
            *args,
            *(f"{k}={v}" for k, v in kwargs.items())  # Note, k should not contain "="
        ])), sep="")

    def table(self) -> OutputTable:
        return MachineTable(self)
    
    def task(self, state: Optional[str], key: Optional[str], **kwargs) -> None:
        self.print_function("task", str(state), str(key), **kwargs)
    
    def finish(self) -> None:
        pass

    def print(self, text: str) -> None:
        self.print_function("print", text)
    
    def prompt(self, password: bool = False) -> Optional[str]:
        self.print_function("prompt", password=str(int(password)))
        try:
            return input("")
        except KeyboardInterrupt:
            return None

class MachineTable(OutputTable):

    def __init__(self, out: MachineOutput) -> None:
        super().__init__()
        self.out = out
    
    def print(self) -> None:
        self.out.print_function("table", str(len(self.rows)))
        for row in self.rows:
            if row is None:
                self.out.print_function("sep")
            else:
                self.out.print_function("row", *row)



"""Entry module for PortableMC Command Line Interface. 

This module implements all (sub)commands of the CLI, the arguments parser, language and 
other utilities are defined in child modules.

**Note that this module, even when no *underscore* "_" is used, should not be considered
as part of the public API.**
"""

from subprocess import Popen
from pathlib import Path
import socket
import sys
import io

from .parse import register_arguments, RootNs, SearchNs, StartNs, LoginNs, LogoutNs, AuthBaseNs, ShowCompletionNs

from .util import format_locale_date, format_time, format_number, anonymize_email
from .output import Output, HumanOutput, MachineOutput, OutputTable
from .lang import get as _, lang

from portablemc.util import LibrarySpecifier
from portablemc.http import HttpError
from portablemc.auth import AuthDatabase, AuthSession, MicrosoftAuthSession, \
    YggdrasilAuthSession, AuthError

from portablemc.standard import Context, Version, VersionManifest, SimpleWatcher, \
    DownloadError, DownloadStartEvent, DownloadProgressEvent, DownloadCompleteEvent, \
    VersionNotFoundError, TooMuchParentsError, FeaturesEvent, JarNotFoundError, \
    JvmNotFoundError, LibraryNotFoundError, \
    VersionLoadingEvent, VersionFetchingEvent, VersionLoadedEvent, \
    JvmLoadingEvent, JvmLoadedEvent, JarFoundEvent, \
    AssetsResolveEvent, LibrariesResolvingEvent, LibrariesResolvedEvent, \
    LoggerFoundEvent, \
    StreamRunner, XmlStreamEvent

from portablemc.fabric import FabricVersion, FabricResolveEvent
from portablemc.forge import ForgeVersion, ForgeResolveEvent, ForgePostProcessingEvent, \
    ForgePostProcessedEvent, ForgeInstallError, _NeoForgeVersion

from typing import cast, Optional, List, Union, Dict, Callable, Any, Tuple


EXIT_OK = 0
EXIT_FAILURE = 1

AUTH_DATABASE_FILE_NAME = "portablemc_auth.json"
MANIFEST_CACHE_FILE_NAME = "portablemc_version_manifest.json"
MICROSOFT_AZURE_APP_ID = "708e91b5-99f8-4a1d-80ec-e746cbb24771"

DEFAULT_JVM_ARGS = [
    "-Xmx2G",
    "-XX:+UnlockExperimentalVMOptions",
    "-XX:+UseG1GC",
    "-XX:G1NewSizePercent=20",
    "-XX:G1ReservePercent=20",
    "-XX:MaxGCPauseMillis=50",
    "-XX:G1HeapRegionSize=32M"
]

CommandHandler = Callable[[Any], Any]
CommandTree = Dict[str, Union[CommandHandler, "CommandTree"]]


def main(args: Optional[List[str]] = None):
    """Main entry point of the CLI. This function parses the input arguments and try to
    find a command handler to dispatch to. These command handlers are specified by the
    `get_command_handlers` function.
    """

    # Force stdout/stderr to use UTF-8 encoding, this reconfigure method 
    # is available for Python 3.7 and onward.
    if isinstance(sys.stdout, io.TextIOWrapper):
        sys.stdout.reconfigure(encoding='utf-8')
    if isinstance(sys.stderr, io.TextIOWrapper):
        sys.stderr.reconfigure(encoding='utf-8')

    parser = register_arguments()
    ns: RootNs = cast(RootNs, parser.parse_args(args or sys.argv[1:]))

    # Setup common objects in the namespace.
    ns.parser = parser
    ns.out = get_output(ns.out_kind)
    ns.context = Context(ns.main_dir, ns.work_dir)
    ns.version_manifest = VersionManifest(ns.context.work_dir / MANIFEST_CACHE_FILE_NAME)
    ns.auth_database = AuthDatabase(ns.context.work_dir / AUTH_DATABASE_FILE_NAME)
    ns.socket_error_tips = []
    socket.setdefaulttimeout(ns.timeout)

    # Find the command handler and run it.
    command_handlers = get_command_handlers()
    command_attr = "subcommand"
    while True:
        command = getattr(ns, command_attr)
        handler = command_handlers.get(command)
        if handler is None:
            parser.print_help()
            sys.exit(EXIT_FAILURE)
        elif callable(handler):
            cmd(handler, ns)
        elif isinstance(handler, dict):
            command_attr = f"{command}_{command_attr}"
            command_handlers = handler
            continue
        sys.exit(EXIT_OK)


def get_output(kind: str) -> Output:
    """Internal function that construct the output depending on its kind.
    The kind is constrained by choices set to the arguments parser.
    """

    if kind == "human-color":
        return HumanOutput(True)
    elif kind == "human":
        return HumanOutput(False)
    elif kind == "machine":
        return MachineOutput()
    else:
        raise ValueError()


def get_command_handlers() -> CommandTree:
    """Internal function returns the tree of command handlers for each subcommand
    of the CLI argument parser.
    """

    return {
        "search": cmd_search,
        "start": cmd_start,
        "login": cmd_login,
        "logout": cmd_logout,
        "show": {
            "about": cmd_show_about,
            "auth": cmd_show_auth,
            "lang": cmd_show_lang,
            "completion": cmd_show_completion,
        },
    }


def cmd(handler: CommandHandler, ns: RootNs):
    """Generic command handler that launch the given handler with the given namespace,
    it handles error in order to pretty print them.
    """
    
    try:
        handler(ns)
        sys.exit(EXIT_OK)
    
    except ValueError as error:
        if len(error.args):
            for i, arg in enumerate(error.args):
                ns.out.task("FAILED" if i == 0 else None, "echo", echo=arg)
                ns.out.finish()
        else:
            ns.out.task("FAILED", "echo", echo="programming error")
            ns.out.finish()
        
        if ns.verbose >= 1:
            import traceback
            traceback.print_exc()
        else:
            ns.out.task("INFO", "suggest_verbose")
            ns.out.finish()
    
    except KeyboardInterrupt:
        ns.out.finish()
        ns.out.task("HALT", "keyboard_interrupt")
        ns.out.finish()
    
    except (OSError, HttpError) as error:

        from urllib.error import URLError
        from ssl import SSLCertVerificationError

        key = "error.os"

        if isinstance(error, HttpError):
            if error.res.status == 0:
                # Status 0 means that network error happened, just forward this error.
                error = error.reason
            else:
                # Unhandled HTTP error code.
                # Here we don't redefine "error" so this will skip following conditions.
                key = "error.http"

        # We are only interested in the reason, URLError is just a wrapper.
        if isinstance(error, URLError):
            error = error.reason

        # More precise errors.
        if isinstance(error, SSLCertVerificationError):
            key = "error.cert"
        elif isinstance(error, (socket.gaierror, socket.timeout)):
            key = "error.socket"
        
        ns.out.task("FAILED", key)
        ns.out.finish()

        if ns.verbose >= 1:
            import traceback
            traceback.print_exc()
        else:
            ns.out.task(None, "echo", echo=repr(error))
            ns.out.finish()
            ns.out.task("INFO", "suggest_verbose")
            ns.out.finish()

        if key == "error.socket":
            for error_tip in ns.socket_error_tips:
                ns.out.task("INFO", f"error.socket.tip.{error_tip}")
                ns.out.finish()
    
    sys.exit(EXIT_FAILURE)


def cmd_search(ns: SearchNs):
    table = ns.out.table()
    cmd_search_handler(ns, ns.kind, table)
    table.print()
    sys.exit(EXIT_OK)

def cmd_search_handler(ns: SearchNs, kind: str, table: OutputTable):
    """Internal function that handles searching a particular kind of search.
    The value of "kind" is constrained by choices in the argument parser.
    """

    search = ns.input

    if kind == "mojang":
        
        table.add(
            _("search.type"),
            _("search.name"),
            _("search.release_date"),
            _("search.flags"))
        table.separator()

        ns.socket_error_tips.append("version_manifest")
 
        if search is not None:
            search, alias = ns.version_manifest.filter_latest(search)
        else:
            alias = False

        for version_data in ns.version_manifest.all_versions():
            version_id = version_data["id"]
            if search is None or (alias and search == version_id) or (not alias and search in version_id):
                version = ns.context.get_version(version_id)
                table.add(
                    version_data["type"], 
                    version_id, 
                    format_locale_date(version_data["releaseTime"]),
                    _("search.flags.local") if version.metadata_exists() else "")
    
    elif kind == "local":

        table.add(
            _("search.name"),
            _("search.last_modified"))
        table.separator()

        search = ns.input
        for version in ns.context.list_versions():
            if search is None or search in version.id:
                table.add(version.id, format_locale_date(version.metadata_file().stat().st_mtime))
    
    elif kind == "forge":

        from ..forge import request_promo_versions
        
        table.add(_("search.name"), _("search.loader_version"))
        table.separator()

        if search is not None:
            ns.socket_error_tips.append("version_manifest")
            search = ns.version_manifest.filter_latest(search)[0]

        for alias, version in request_promo_versions().items():
            if search is None or search in alias:
                table.add(alias, version)

    elif kind in ("fabric", "quilt", "legacyfabric", "babric"):

        from ..fabric import FABRIC_API, QUILT_API, LEGACYFABRIC_API, BABRIC_API

        table.add(_("search.loader_version"), _("search.flags"))
        table.separator()

        if kind == "fabric":
            api = FABRIC_API
        elif kind == "quilt":
            api = QUILT_API
        elif kind == "legacyfabric":
            api = LEGACYFABRIC_API
        elif kind == "babric":
            api = BABRIC_API
        
        for loader in api._request_loaders():
            if search is None or search in loader.version:
                table.add(loader.version, _("search.flags.stable") if loader.stable else "")

    else:
        raise ValueError()


def cmd_start(ns: StartNs):

    version_parts = ns.version.split(":")
    
    # If no split, the kind of version is "standard": parts have at least 2 elements.
    if len(version_parts) == 1:
        version_parts = ["standard", version_parts[0]]

    # No handler means that the format is invalid.
    version = cmd_start_handler(ns, version_parts[0], version_parts[1:])
    if version is None:
        format_key = f"args.start.version.{version_parts[0]}"
        if format_key not in lang:
            ns.out.task("FAILED", "start.version.invalid_id_unknown_kind", kind=version_parts[0])
        else:
            ns.out.task("FAILED", "start.version.invalid_id", expected=_(format_key))
        ns.out.finish()
        sys.exit(EXIT_FAILURE)

    version.manifest = ns.version_manifest
    version.disable_multiplayer = ns.disable_mp
    version.disable_chat = ns.disable_chat
    version.demo = ns.demo
    version.resolution = ns.resolution
    version.jvm_path = ns.jvm

    if ns.server is not None:
        version.set_quick_play_multiplayer(ns.server, ns.server_port or 25565)

    if ns.no_fix:
        version.fixes.clear()
    
    if ns.lwjgl is not None:
        version.fixes[Version.FIX_LWJGL] = ns.lwjgl

    if ns.login is not None:
        version.auth_session = prompt_authenticate(ns, ns.login, not ns.temp_login, ns.auth_anonymize)
        if version.auth_session is None:
            sys.exit(EXIT_FAILURE)
    else:
        version.set_auth_offline(ns.username, ns.uuid)

    # Excluded libraries
    if ns.exclude_lib is not None:

        exclude_filters = ns.exclude_lib
        def filter_libraries(libs: Dict[LibrarySpecifier, Any]) -> None:
            # Here the complexity is terrible, but I guess it's acceptable?
            to_del = []
            unused_filters = set(exclude_filters)
            for spec in libs.keys():
                for spec_filter in exclude_filters:
                    if spec_filter.matches(spec):
                        unused_filters.remove(spec_filter)
                        to_del.append(spec)
                        break
            # Finally delete selected specifiers
            for spec in to_del:
                del libs[spec]
                if ns.verbose >= 1:
                    ns.out.task("INFO", "start.libraries.excluded", spec=str(spec))
                    ns.out.finish()
            # Inform the user of unused filters
            for unused_filter in unused_filters:
                ns.out.task("WARN", "start.libraries.unused_filter", filter=str(unused_filter))
                ns.out.finish()
        
        version.libraries_filters.append(filter_libraries)

    try:

        env = version.install(watcher=StartWatcher(ns))

        if ns.verbose >= 1:
            for fix, fix_value in env.fixes.items():
                ns.out.task("INFO", f"start.fix.{fix}", value=fix_value)
                ns.out.finish()

        # Included binaries
        if ns.include_bin is not None:
            for bin_path in ns.include_bin:
                if not bin_path.is_file():
                    ns.out.task("FAILED", "start.additional_binary_not_found", path=bin_path)
                    ns.out.finish()
                    sys.exit(EXIT_FAILURE)
                env.native_libs.append(bin_path)
            
        # Extend JVM arguments with given arguments, or defaults
        if ns.jvm_args is None:
            env.jvm_args.extend(DEFAULT_JVM_ARGS)
        elif len(ns.jvm_args):
            env.jvm_args.extend(ns.jvm_args.split())

        # This CliRunner will abort running if in dry mode.
        env.run(CliRunner(ns))
        sys.exit(EXIT_OK)
    
    except VersionNotFoundError as error:
        ns.out.task("FAILED", "start.version.not_found", version=error.version)
        ns.out.finish()
    
    except TooMuchParentsError as error:
        ns.out.task("FAILED", "start.version.too_much_parents")
        ns.out.finish()
        ns.out.task(None, "echo", echo=", ".join(error.versions))
        ns.out.finish()

    except JarNotFoundError as error:
        ns.out.task("FAILED", "start.jar.not_found")
        ns.out.finish()

    except JvmNotFoundError as error:
        ns.out.task("FAILED", f"start.jvm.not_found_error.{error.code}")
        ns.out.finish()
    
    except LibraryNotFoundError as error:
        ns.out.task("FAILED", f"start.libraries.not_found_error", spec=str(error.lib))
        ns.out.finish()
    
    except ForgeInstallError as error:
        ns.out.task("FAILED", f"start.forge.install_error.{error.code}")
        ns.out.finish()

    except DownloadError as error:
        ns.out.task("FAILED", None)
        ns.out.finish()
        for entry, code, _origin in error.errors:
            ns.out.task(None, "download.error", name=entry.url, message=_(f"download.error.{code}"))
            ns.out.finish()
    
    sys.exit(EXIT_FAILURE)

def cmd_start_handler(ns: StartNs, kind: str, parts: List[str]) -> Optional[Version]:
    """This function handles particular kind of versions. If this function successfully
    decodes, the corresponding version should be returned. The global version's format 
    being parsed is <kind>[:<part>..].

    The parts list contains at least one element.

    This function returns false if parsing fail, in such case the expected format is
    printed out to the user on output (language key: "args.start.version.<kind>").
    """

    version = parts[0] or "release"
    ns.socket_error_tips.append("version_manifest")
    
    if ns.verbose >= 1:
        ns.out.task("INFO", "start.global_version", kind=kind, version=version, remaining=" ".join(parts[1:]))
        ns.out.finish()

    if kind == "standard":
        if len(parts) != 1:
            return None
        return Version(version, context=ns.context)
    
    elif kind in ("fabric", "quilt", "legacyfabric", "babric"):

        if len(parts) > 2:
            return None
        
        # Legacy fabric has a special case because it will never be supported for 
        # versions past 1.13.2, it is not made for latest release version.
        if version == "release":
            if kind == "legacyfabric":
                version = "1.13.2"
            elif kind == "babric":
                version = "b1.7.3"
        
        if kind == "fabric":
            constructor = FabricVersion.with_fabric
            prefix = ns.fabric_prefix
        elif kind == "quilt":
            constructor = FabricVersion.with_quilt
            prefix = ns.quilt_prefix
        elif kind == "legacyfabric":
            constructor = FabricVersion._with_legacyfabric
            prefix = ns.legacyfabric_prefix
        elif kind == "babric":
            constructor = FabricVersion._with_babric
            prefix = ns.babric_prefix
        
        if len(parts) != 2:
            ns.socket_error_tips.append(f"{kind}_loader_version")

        return constructor(version, parts[1] if len(parts) == 2 else None, context=ns.context, prefix=prefix)
    
    elif kind in ("forge", "neoforge"):
        if len(parts) != 1:
            return None
        constructor = ForgeVersion if kind == "forge" else _NeoForgeVersion
        prefix = ns.forge_prefix if kind == "forge" else ns.neoforge_prefix
        return constructor(version, context=ns.context, prefix=prefix)
    
    else:
        return None


def cmd_login(ns: LoginNs):
    session = prompt_authenticate(ns, ns.email_or_username, True)
    if session is not None:
        ns.out.task("INFO", "login.tip.remember_start_login", email=ns.email_or_username)
        ns.out.finish()
    sys.exit(EXIT_FAILURE if session is None else EXIT_OK)


def cmd_logout(ns: LogoutNs):

    session_class = {
        "microsoft": MicrosoftAuthSession,
        "yggdrasil": YggdrasilAuthSession,
    }[ns.auth_service]

    ns.out.task("", f"logout.{ns.auth_service}.pending", email=ns.email_or_username)
    ns.auth_database.load()
    session = ns.auth_database.remove(ns.email_or_username, session_class)
    if session is not None:
        session.invalidate()
        ns.auth_database.save()
        ns.out.task("OK", "logout.success", email=ns.email_or_username)
        ns.out.finish()
        sys.exit(EXIT_OK)
    else:
        ns.out.task("FAILED", "logout.unknown_session", email=ns.email_or_username)
        ns.out.finish()
        sys.exit(EXIT_FAILURE)


def cmd_show_about(ns: RootNs):
    
    from .. import LAUNCHER_VERSION, LAUNCHER_AUTHORS, LAUNCHER_URL, LAUNCHER_COPYRIGHT

    print(f"Version: {LAUNCHER_VERSION}")
    print(f"Authors: {', '.join(LAUNCHER_AUTHORS)}")
    print(f"Website: {LAUNCHER_URL}")
    print(f"License: {LAUNCHER_COPYRIGHT}")
    print( "         This program comes with ABSOLUTELY NO WARRANTY. This is free software,")
    print( "         and you are welcome to redistribute it under certain conditions.")
    print( "         See <https://www.gnu.org/licenses/gpl-3.0.html>.")


def cmd_show_auth(ns: RootNs):

    ns.auth_database.load()
    table = ns.out.table()

     # Intentionally not i18n for now because used for debug purpose.
    table.add("Type", "Email", "Username", "UUID")
    table.separator()

    for auth_type, auth_type_sessions in ns.auth_database.sessions.items():
        for email, sess in auth_type_sessions.items():
            table.add(auth_type, email, sess.username, sess.uuid)
    
    table.print()


def cmd_show_lang(ns: RootNs):

    from .lang import lang

    table = ns.out.table()

     # Intentionally not i18n for now because used for debug purpose.
    table.add("Key", "Message")
    table.separator()

    for key, msg in lang.items():
        table.add(key, msg)

    table.print()


def cmd_show_completion(ns: ShowCompletionNs):
    
    from .complete import gen_zsh_completion, gen_bash_completion

    if ns.shell == "zsh":
        content = gen_zsh_completion(ns.parser)
    elif ns.shell == "bash":
        content = gen_bash_completion(ns.parser)
    else:
        raise RuntimeError
    
    print(content, end="")


def prompt_authenticate(ns: AuthBaseNs, email: str, caching: bool, anonymise: bool = False) -> Optional[AuthSession]:
    """Prompt the user to login using the given email (or legacy username) for specific 
    service (Microsoft or Yggdrasil) and return the :class:`AuthSession` if successful, 
    None otherwise. This function handles task printing and all exceptions are caught 
    internally.
    """

    service = ns.auth_service

    session_class = {
        "microsoft": MicrosoftAuthSession,
        "yggdrasil": YggdrasilAuthSession,
    }[service]

    ns.auth_database.load()

    if service == "yggdrasil":
        ns.out.task("WARN", "auth.yggdrasil.deprecated")
        ns.out.finish()

    task_text = f"auth.{service}"
    email_text = anonymize_email(email) if anonymise else email

    ns.out.task("..", task_text, email=email_text)

    session = ns.auth_database.get(email, session_class)
    if session is not None:
        try:
            
            if not session.validate():
                ns.out.task("..", "auth.refreshing")
                session.refresh()
                ns.auth_database.save()
                ns.out.task("OK", "auth.refreshed", email=email_text)
            else:
                ns.out.task("OK", "auth.validated", email=email_text)

            ns.out.finish()
            return session
        
        except AuthError as error:
            ns.out.task("FAILED", None)
            ns.out.task(None, "auth.error", message=str(error))
            ns.out.finish()

    ns.out.task("..", task_text, email=email_text)

    try:

        if service == "microsoft":
            session = prompt_microsoft_authenticate(ns, email)
        else:
            session = prompt_yggdrasil_authenticate(ns, email)
        
    except AuthError as error:
        ns.out.task("FAILED", "auth.error", message=str(error))
        ns.out.finish()
        return None

    if session is None:
        return None
    if caching:
        ns.out.task("..", "auth.caching")
        ns.auth_database.put(email, session)
        ns.auth_database.save()
    
    ns.out.task("OK", "auth.logged_in", email=email_text)
    ns.out.finish()

    return session


def prompt_yggdrasil_authenticate(ns: RootNs, email_or_username: str) -> Optional[YggdrasilAuthSession]:
    ns.out.finish()
    ns.out.task("..", "auth.yggdrasil.enter_password")
    password = ns.out.prompt(password=True)
    if password is None:
        ns.out.task("FAILED", "cancelled")
        ns.out.finish()
        return None
    else:
        return YggdrasilAuthSession.authenticate(ns.auth_database.get_client_id(), email_or_username, password)


def prompt_microsoft_authenticate(ns: AuthBaseNs, email: str) -> Optional[MicrosoftAuthSession]:

    from .. import LAUNCHER_NAME, LAUNCHER_VERSION
    from http.server import HTTPServer, BaseHTTPRequestHandler
    from uuid import uuid4
    import urllib.parse
    import webbrowser

    nonce = uuid4().hex
    app_id = MICROSOFT_AZURE_APP_ID
    redirect_uri = "https://www.theorozier.fr/portablemc/auth"

    def gen_auth_url(state: str) -> str:
        return "https://login.live.com/oauth20_authorize.srf?{}".format(urllib.parse.urlencode({
            "client_id": app_id,
            "redirect_uri": redirect_uri,
            "response_type": "code id_token",
            "scope": "xboxlive.signin offline_access openid email",
            "login_hint": email,
            "nonce": nonce,
            "state": state,
            "prompt": "login",
            "response_mode": "fragment"
        }))
    
    auth_query = None
    
    if not ns.auth_no_browser:

        class AuthServer(HTTPServer):

            def __init__(self):
                super().__init__(("127.0.0.1", 0), RequestHandler)
                self.timeout = 0.5
                self.ms_auth_query: Optional[str] = None

        class RequestHandler(BaseHTTPRequestHandler):

            server_version = f"{LAUNCHER_NAME}/{LAUNCHER_VERSION}"

            def __init__(self, request, client_address: Tuple[str, int], auth_server: AuthServer) -> None:
                super().__init__(request, client_address, auth_server)

            def log_message(self, _format: str, *args: Any):
                return

            def do_GET(self):
                parsed = urllib.parse.urlparse(self.path)
                if parsed.path in ("", "/"):
                    cast(AuthServer, self.server).ms_auth_query = parsed.query
                    self.send_response(200)
                else:
                    self.send_response(404)
                self.send_header("Access-Control-Allow-Origin", "*")
                self.end_headers()
                self.wfile.flush()
    
        # We start be creating the authentication server, at this point we don't start it 
        # but we have allocated a free port.
        with AuthServer() as server:

            # First try opening the authentication page with the local webbrowser
            if webbrowser.open(gen_auth_url(f"port:{server.server_port}")):
                # If successfully opened the browser, we actually start the web server.
                ns.out.task("..", "auth.microsoft.opening_browser_and_listening")
                try:
                    while server.ms_auth_query is None:
                        server.handle_request()
                except KeyboardInterrupt:
                    pass
            
            if server.ms_auth_query is None:
                ns.out.finish()
                ns.out.task("FAILED", "auth.microsoft.failed_to_authenticate")
                ns.out.finish()
                return None

            auth_query = server.ms_auth_query

    # If here we have code or id token none, it means that no web browser has been opened.
    # So we want to print the URL auth URL so the user can try manually.
    if auth_query is None:

        ns.out.task("INFO", "auth.microsoft.no_browser_fallback")
        ns.out.finish()
        ns.out.print(gen_auth_url("") + "\n")

        ns.out.task("..", "auth.microsoft.no_browser_code")
        auth_query = ns.out.prompt()
        if auth_query is None:
            ns.out.finish()
            ns.out.task("FAILED", "auth.microsoft.failed_to_authenticate")
            ns.out.finish()
            return None
        else:
            auth_query = auth_query.strip()

    qs = urllib.parse.parse_qs(auth_query)

    if "code" in qs and "id_token" in qs:

        ns.out.task("..", "auth.microsoft.processing")
        id_token = qs["id_token"][0]
        code = qs["code"][0]

        if not MicrosoftAuthSession.check_token_id(id_token, email, nonce):
            ns.out.finish()
            ns.out.task("FAILED", "auth.microsoft.incoherent_data")
            ns.out.finish()
            return None

        return MicrosoftAuthSession.authenticate(ns.auth_database.get_client_id(), app_id, code, redirect_uri)

    else:
        ns.out.finish()
        ns.out.task("FAILED", "auth.microsoft.failed_to_authenticate")
        ns.out.finish()
        return None

    
class StartWatcher(SimpleWatcher):

    def __init__(self, ns: RootNs) -> None:

        def progress_task(key: str, **kwargs) -> None:
            ns.out.task("..", key, **kwargs)

        def finish_task(key: str, **kwargs) -> None:
            ns.out.task("OK", key, **kwargs)
            ns.out.finish()
        
        def features(e: FeaturesEvent) -> None:
            if ns.verbose >= 1:
                ns.out.task("INFO", "start.features", features=", ".join(e.features))
                ns.out.finish()
        
        def assets_resolve(e: AssetsResolveEvent) -> None:
            if e.count is None:
                ns.out.task("..", "start.assets.resolving", index_version=e.index_version)
            else:
                ns.out.task("OK", "start.assets.resolved", index_version=e.index_version, count=e.count)
                ns.out.finish()

        def libraries_resolved(e: LibrariesResolvedEvent) -> None:
            ns.out.task("OK", "start.libraries.resolved", class_libs_count=e.class_libs_count, native_libs_count=e.native_libs_count)
            ns.out.finish()

        def fabric_resolve(e: FabricResolveEvent) -> None:
            if e.loader_version is None:
                ns.out.task("..", "start.fabric.resolving", api=e.api.name, vanilla_version=e.vanilla_version)
            else:
                ns.out.task("OK", "start.fabric.resolved", api=e.api.name, loader_version=e.loader_version, vanilla_version=e.vanilla_version)
                ns.out.finish()
        
        def forge_resolve(e: ForgeResolveEvent) -> None:
            if e.alias:
                ns.out.task("..", "start.forge.resolving", api=e._api, version=e.forge_version)
            else:
                ns.out.task("OK", "start.forge.resolved", api=e._api, version=e.forge_version)
                ns.out.finish()

        super().__init__({
            VersionLoadingEvent: lambda e: progress_task("start.version.loading", version=e.version),
            VersionFetchingEvent: lambda e: progress_task("start.version.fetching", version=e.version),
            VersionLoadedEvent: lambda e: finish_task("start.version.loaded.fetched" if e.fetched else "start.version.loaded", version=e.version),
            FeaturesEvent: features,
            JvmLoadingEvent: lambda e: progress_task("start.jvm.loading"),
            JvmLoadedEvent: lambda e: finish_task(f"start.jvm.loaded.{e.kind}", version=e.version or ""),
            JarFoundEvent: lambda e: finish_task("start.jar.found"),
            AssetsResolveEvent: assets_resolve,
            LibrariesResolvingEvent: lambda e: progress_task("start.libraries.resolving"),
            LibrariesResolvedEvent: libraries_resolved,
            LoggerFoundEvent: lambda e: finish_task("start.logger.found", version=e.version),
            FabricResolveEvent: fabric_resolve,
            ForgeResolveEvent: forge_resolve,
            ForgePostProcessingEvent: lambda e: progress_task("start.forge.post_processing", task=e.task),
            ForgePostProcessedEvent: lambda e: finish_task("start.forge.post_processed"),
            DownloadStartEvent: self.download_start,
            DownloadProgressEvent: self.download_progress,
            DownloadCompleteEvent: self.download_complete,
        })
            
        self.ns = ns
        self.entries_count: int
        self.total_size: int
        self.speeds: List[float]
        self.sizes: List[int]
        self.size = 0

    def download_start(self, e: DownloadStartEvent):

        if self.ns.verbose:
            self.ns.out.task("INFO", "download.threads_count", count=e.threads_count)
            self.ns.out.finish()

        self.entries_count = e.entries_count
        self.total_size = e.size
        self.speeds = [0.0] * e.threads_count
        self.sizes = [0] * e.threads_count
        self.size = 0
        self.ns.out.task("..", "download.start")

    def download_progress(self, e: DownloadProgressEvent) -> None:

        self.speeds[e.thread_id] = e.speed
        self.sizes[e.thread_id] = e.size

        speed = sum(self.speeds)
        total_count = str(self.entries_count)
        count = f"{e.count:{len(total_count)}}"
        
        self.ns.out.task("..", "download.progress", 
            count=count,
            total_count=total_count,
            size=f"{format_number(self.size + sum(self.sizes))}B",
            speed=f"{format_number(speed)}B/s")

        if e.done:
            self.size += e.size

    def download_complete(self, e: DownloadCompleteEvent) -> None:
        self.ns.out.task("OK", None)
        self.ns.out.finish()


class CliRunner(StreamRunner):

    def __init__(self, ns: StartNs) -> None:
        super().__init__()
        self.ns = ns

    def process_create(self, args: List[str], work_dir: Path) -> Optional[Popen]:
        
        if not self.ns.dry or self.ns.verbose >= 2:
            self.ns.out.print("\n")

        if self.ns.verbose >= 2:
            self.ns.out.print(" ".join(args) + "\n")

        if self.ns.dry:
            return None
        
        return super().process_create(args, work_dir)

    def process_stream_event(self, event: Any) -> None:

        out = self.ns.out

        if isinstance(event, XmlStreamEvent):
            time = format_time(event.time)
            out.print(f"[{time}] [{event.thread}] [{event.level}] {event.message}\n")
            if event.throwable is not None:
                out.print(f"{event.throwable.rstrip()}\n")
        else:
            out.print(str(event))



"""Global utilities for the CLI.
"""

from datetime import datetime

from portablemc.util import LibrarySpecifier, from_iso_date

from typing import Optional, Union


def format_locale_date(raw: Union[str, float]) -> str:
    if isinstance(raw, float):
        return datetime.fromtimestamp(raw).strftime("%c")
    else:
        return from_iso_date(str(raw)).strftime("%c")


def format_time(timestamp: float) -> str:
    """Format the given timestamp (in seconds) into hh:mm:ss format.
    """
    return datetime.fromtimestamp(timestamp).strftime("%H:%M:%S")


def format_number(n: float) -> str:
    """Return a number with suffix k, M, G or nothing. 
    The string is at most 7 chars unless the size exceed 1 T.
    """
    if n < 1000:
        return f"{int(n)} "
    elif n < 1000000:
        return f"{(int(n / 100) / 10):.1f} k"
    elif n < 1000000000:
        return f"{(int(n / 100000) / 10):.1f} M"
    else:
        return f"{(int(n / 100000000) / 10):.1f} G"


def format_duration(n: float) -> str:
    """Return a duration with proper suffix s, m, h.
    """
    if n < 60:
        return f"{int(n)} s"
    elif n < 3600:
        return f"{int(n / 60)} m"
    else:
        return f"{int(n / 3600)} h"
    

def anonymize_email(email: str) -> str:
    """Return a visually anonymized email.
    """

    def anonymize_part(email_part: str) -> str:
        return f"{email_part[0]}{'*' * (len(email_part) - 2)}{email_part[-1]}"
    
    parts = []

    for i, part in enumerate(email.split("@", maxsplit=1)):
        if i == 0:
            parts.append(anonymize_part(part))
        else:
            parts.append(".".join((anonymize_part(server_part) if j == 0 else server_part for j, server_part in enumerate(part.split(".", maxsplit=1)))))
    
    return "@".join(parts)


class LibrarySpecifierFilter:
    """A filter for library specifier, used with the start command to exclude some 
    libraries.
    """
    
    __slots__ = "artifact", "version", "classifier"

    def __init__(self, artifact: str, version: Optional[str], classifier: Optional[str]):
        self.artifact = artifact
        self.version = version
        self.classifier = classifier
    
    @classmethod
    def from_str(cls, s: str) -> "LibrarySpecifierFilter":

        parts = s.split(":")
        if len(parts) > 3:
            raise ValueError("Invalid parts count")

        return cls(parts[0], 
                   parts[1] or None if len(parts) >= 2 else None, 
                   parts[2] or None if len(parts) >= 3 else None)

    def matches(self, spec: LibrarySpecifier) -> bool:
        return self.artifact == spec.artifact \
            and (self.version is None or self.version == spec.version) \
            and (self.classifier is None or (spec.classifier or "").startswith(self.classifier))

    def __str__(self) -> str:
        return f"{self.artifact}:{self.version or ''}" + ("" if self.classifier is None else f":{self.classifier}")

    def __repr__(self) -> str:
        return f"<LibrarySpecifierFilter {self}>"


"""Module providing tasks for launching forge mod loader versions.

The NeoForge support is still unstable API.
"""

from urllib import parse as url_parse
from zipfile import ZipFile
from pathlib import Path
from io import BytesIO
import subprocess
import shutil
import json
import os

from .standard import parse_download_entry, LIBRARIES_URL, \
    Context, VersionHandle, Version, Watcher, VersionNotFoundError

from .util import calc_input_sha1, LibrarySpecifier
from .http import http_request, HttpError

from typing import Dict, Optional, List, Tuple


_FORGE_REPO_URL = "https://maven.minecraftforge.net/"
_FORGE_GROUP = "net.minecraftforge"
_FORGE_ARTIFACT = "forge"

_NEO_FORGE_REPO_URL = "https://maven.neoforged.net/releases/"
_NEO_FORGE_GROUP = "net.neoforged"
_NEO_FORGE_ARTIFACT = "neoforge"


class ForgeVersion(Version):
    
    def __init__(self, forge_version: str = "release", *,
        context: Optional[Context] = None,
        prefix: str = "forge",
    ) -> None:

        super().__init__("", context=context)  # Do not give a root version for now.

        self.forge_version = forge_version
        self.prefix = prefix
        
        # This is set when version is resolved.
        self._forge_repo_url: Optional[str] = None
        self._forge_installer_spec: Optional[LibrarySpecifier] = None

        self._forge_post_info: Optional[_ForgePostInfo] = None
    
    def _resolve_version(self, watcher: Watcher) -> None:
        
        # Maybe "release" or "snapshot", we process this first.
        self.forge_version = self.manifest.filter_latest(self.forge_version)[0]

        # If no alias is specified, with add recommended.
        if "-" not in self.forge_version:
            self.forge_version = f"{self.forge_version}-recommended"

        # Now, if the specified version is an alias, we resolve it.
        if self.forge_version.endswith(("-latest", "-recommended")):

            # Split the version, used later.
            alias_version, alias = self.forge_version.rsplit("-", maxsplit=1)
            watcher.handle(ForgeResolveEvent(self.forge_version, True))

            # Try to get loader from promo versions.
            promo_versions = request_promo_versions()
            loader_version = promo_versions.get(self.forge_version)

            # If we can't find the load version, just try to other alias (issue #189).
            if loader_version is None:
                alias = { "latest": "recommended", "recommended": "latest" }[alias]
                self.forge_version = f"{alias_version}-{alias}"
                watcher.handle(ForgeResolveEvent(self.forge_version, True))
                loader_version = promo_versions.get(self.forge_version)

            if loader_version is None:
                raise VersionNotFoundError(f"{self.prefix}-{alias_version}-???")

            self.forge_version = f"{alias_version}-{loader_version}"
            watcher.handle(ForgeResolveEvent(self.forge_version, False))
        
        # Finally define the full version id.
        self.version = f"{self.prefix}-{self.forge_version}"
        self._forge_repo_url = _FORGE_REPO_URL
        self._forge_installer_spec = LibrarySpecifier(_FORGE_GROUP, _FORGE_ARTIFACT, self.forge_version, classifier="installer")

    def _load_version(self, version: VersionHandle, watcher: Watcher) -> bool:
        if version.id == self.version:
            return version.read_metadata_file()
        else:
            return super()._load_version(version, watcher)

    def _fetch_version(self, version: VersionHandle, watcher: Watcher) -> None:

        if version.id != self.version:
            return super()._fetch_version(version, watcher)
        
        # Must have been set in _resolve_version
        assert self._forge_repo_url is not None
        assert self._forge_installer_spec is not None

        # Extract the game version from the forge version, we'll use
        # it to add suffix to find the right forge version if needed.
        game_version = self.forge_version.split("-", 1)[0]

        # For some older game versions, some odd suffixes were used 
        # for the version scheme.
        suffixes = [""] + {
            "1.11":     ["-1.11.x"],
            "1.10.2":   ["-1.10.0"],
            "1.10":     ["-1.10.0"],
            "1.9.4":    ["-1.9.4"],
            "1.9":      ["-1.9.0", "-1.9"],
            "1.8.9":    ["-1.8.9"],
            "1.8.8":    ["-1.8.8"],
            "1.8":      ["-1.8"],
            "1.7.10":   ["-1.7.10", "-1710ls", "-new"],
            "1.7.2":    ["-mc172"],
        }.get(game_version, [])

        # Iterate suffix and find the first install JAR that works.
        install_jar = None
        original_version = self._forge_installer_spec.version
        for suffix in suffixes:
            try:
                # Apply the suffix before request...
                self._forge_installer_spec.version = f"{original_version}{suffix}"
                install_jar_url = f"{self._forge_repo_url}{self._forge_installer_spec.file_path()}"
                install_jar_res = res = http_request("GET", install_jar_url, accept="application/java-archive")
                install_jar = ZipFile(BytesIO(install_jar_res.data))
                break
            except HttpError as error:
                if error.res.status != 404:
                    raise
                # Silently ignore if the file was not found or forbidden.
                pass
        
        if install_jar is None:
            raise VersionNotFoundError(version.id)
        
        with install_jar:

            # The install profiles comes in multiples forms:
            # 
            # >= 1.12.2-14.23.5.2851
            #  There are two files, 'install_profile.json' which 
            #  contains processors and shared data, and `version.json`
            #  which is the raw version meta to be fetched.
            #
            # <= 1.12.2-14.23.5.2847
            #  There is only an 'install_profile.json' with the version
            #  meta stored in 'versionInfo' object. Each library have
            #  two keys 'serverreq' and 'clientreq' that should be
            #  removed when the profile is returned.

            try:
                info = install_jar.getinfo("install_profile.json")
                with install_jar.open(info) as fp:
                    install_profile = json.load(fp)
            except KeyError:
                raise ForgeInstallError(self.forge_version, ForgeInstallError.INSTALL_PROFILE_NOT_FOUND)

            # print(f"{json.dumps(install_profile, indent='  ')}")

            if "json" in install_profile:  # Forge versions since 1.12.2-14.23.5.2851
                
                info = install_jar.getinfo(install_profile["json"].lstrip("/"))
                with install_jar.open(info) as fp:
                    version.metadata = json.load(fp)

                # We use the bin directory if there is a need to extract temporary files.
                post_info = _ForgePostInfo(self.context.gen_bin_dir())

                # Parse processors
                for i, processor in enumerate(install_profile["processors"]):

                    processor_sides = processor.get("sides", [])
                    if not isinstance(processor_sides, list):
                        raise ValueError(f"forge profile: /json/processors/{i}/sides must be an array")

                    if len(processor_sides) and "client" not in processor_sides:
                        continue

                    processor_jar_name = processor.get("jar")
                    if not isinstance(processor_jar_name, str):
                        raise ValueError(f"forge profile: /json/processors/{i}/jar must be a string")

                    processor_spec = LibrarySpecifier.from_str(processor_jar_name)

                    post_info.processors.append(_ForgePostProcessor(
                        processor_spec,
                        [LibrarySpecifier.from_str(raw_spec) for raw_spec in processor.get("classpath", [])],
                        processor.get("args", []),
                        processor.get("outputs", {})
                    ))

                # Some early (still modern) installers (<= 1.16.5) embed the forge JAR,
                # we need to extract it given its path.
                forge_spec_raw = install_profile.get("path")
                if forge_spec_raw is not None:
                    lib_spec = LibrarySpecifier.from_str(forge_spec_raw)
                    lib_path = self.context.libraries_dir / lib_spec.file_path()
                    zip_extract_file(install_jar, f"maven/{lib_spec.file_path()}", lib_path)

                # We fetch all libraries used to build artifacts, and we store each path 
                # to each library here. These install profile libraries are only used for
                # building, and will be used in finalize task.
                for i, install_lib in enumerate(install_profile["libraries"]):

                    lib_name = install_lib["name"]
                    lib_spec = LibrarySpecifier.from_str(lib_name)
                    lib_artifact = install_lib["downloads"]["artifact"]
                    lib_path = self.context.libraries_dir / lib_spec.file_path()

                    # Ignore the library if it has already been specified, has been seen
                    # in neoforge installer...
                    if lib_spec in post_info.libraries:
                        continue
                    post_info.libraries[lib_spec] = lib_path
                    # print(lib_spec, lib_path)
                    
                    if len(lib_artifact["url"]):
                        self._dl.add(parse_download_entry(lib_artifact, lib_path, "forge profile: /json/libraries/"), verify=True)
                    else:
                        # The lib should be stored inside the JAR file, under maven/ directory.
                        zip_extract_file(install_jar, f"maven/{lib_spec.file_path()}", lib_path)

                # Just keep the 'client' values.
                install_data = install_profile["data"]
                if isinstance(install_data, dict):
                    for data_key, data_val in install_data.items():

                        data_val = str(data_val["client"])

                        # Refer to a file inside the JAR file.
                        if data_val.startswith("/"):
                            dst_path = post_info.tmp_dir / data_val[1:]
                            zip_extract_file(install_jar, data_val[1:], dst_path)
                            data_val = str(dst_path.absolute())  # Replace by the path of extracted file.

                        post_info.variables[data_key] = data_val
                
                self._forge_post_info = post_info

            else:  # Forge versions before 1.12.2-14.23.5.2847

                version.metadata = install_profile.get("versionInfo")
                if not isinstance(version.metadata, dict):
                    raise ForgeInstallError(self.forge_version, ForgeInstallError.VERSION_METADATA_NOT_FOUND)
                
                # Older versions have non standard keys for libraries.
                for version_lib in version.metadata["libraries"]:
                    if "serverreq" in version_lib:
                        del version_lib["serverreq"]
                    if "clientreq" in version_lib:
                        del version_lib["clientreq"]
                    if "checksums" in version_lib:
                        del version_lib["checksums"]
                    # Older versions used to require libraries that are no longer installed
                    # by parent versions, therefore it's required to add url if not 
                    # provided, pointing to maven central repository, for downloading.
                    if not version_lib.get("url"):
                        version_lib["url"] = LIBRARIES_URL
                
                # Old version (<= 1.6.4) of forge are broken, even on official launcher.
                # So we fix them by manually adding the correct inherited version.
                if "inheritsFrom" not in version.metadata:
                    version.metadata["inheritsFrom"] = install_profile["install"]["minecraft"]

                # For "old" installers, that have an "install" section.
                jar_entry_path = install_profile["install"]["filePath"]
                jar_spec = LibrarySpecifier.from_str(install_profile["install"]["path"])
                
                # Here we copy the forge jar stored to libraries.
                jar_path = self.context.libraries_dir / jar_spec.file_path()
                zip_extract_file(install_jar, jar_entry_path, jar_path)

        version.metadata["id"] = version.id
        version.write_metadata_file()
    
    def _resolve_jar(self, watcher: Watcher) -> None:
        super()._resolve_jar(watcher)
        self._finalize_forge(watcher)
    
    def _finalize_forge(self, watcher: Watcher) -> None:
        try:
            self._finalize_forge_internal(watcher)
        except:
            # We just intercept errors and remove the version metadata in case of errors,
            # this allows us to re-run the whole install on next attempt.
            try:
                self._hierarchy[0].metadata_file().unlink()
            except FileNotFoundError:
                pass  # Not a problem if the file isn't present.
            raise

    def _finalize_forge_internal(self, watcher: Watcher) -> None:
        """This step finalize the forge installation, after both JVM and version's JAR
        files has been resolved. This is not always used, it depends on installer's
        version.
        """

        info = self._forge_post_info
        if info is None:
            return
        
        assert self._jvm_path is not None, "_resolve_jvm(...) missing"
        assert self._jar_path is not None, "_resolve_jar(...) missing"
        
        # Download JVM files and version's JAR. This is used for finalization.
        self._download(watcher)

        # Additional missing variables, the version's jar file is the same as the vanilla
        # one, so we use its path.
        info.variables["SIDE"] = "client"
        info.variables["MINECRAFT_JAR"] = str(self._jar_path.absolute())

        def replace_install_args(txt: str) -> str:
            txt = txt.format_map(info.variables)
            # Replace the pattern [lib name] with lib path.
            if txt[0] == "[" and txt[-1] == "]":
                spec = LibrarySpecifier.from_str(txt[1:-1])
                txt = str((self.context.libraries_dir / spec.file_path()).absolute())
            elif txt[0] == "'" and txt[-1] == "'":
                txt = txt[1:-1]
            return txt

        for processor in info.processors:

            # Extract the main-class from manifest. Required because we cannot use 
            # both -cp and -jar.
            jar_path = info.libraries[processor.spec].absolute()
            main_class = None
            with ZipFile(jar_path) as jar_fp:
                with jar_fp.open("META-INF/MANIFEST.MF") as manifest_fp:
                    for manifest_line in manifest_fp.readlines():
                        if manifest_line.startswith(b"Main-Class: "):
                            main_class = manifest_line[12:].decode().strip()
                            break
            
            if main_class is None:
                raise ValueError(f"cannot find main class in {jar_path}")

            # Try to find the task name in the arguments, just for information purpose.
            if len(processor.args) >= 2 and processor.args[0] == "--task":
                task = processor.args[1].lower()
            else:
                task = {
                    "jarsplitter": "split_jar",
                    "ForgeAutoRenamingTool": "forge_auto_renaming",
                    "binarypatcher": "patch_binary",
                    "SpecialSource": "special_source_renaming",
                }.get(processor.spec.artifact, f"unknown({processor.spec})")

            # Compute the full arguments list.
            args = [
                str(self._jvm_path.absolute()),
                "-cp", os.pathsep.join([str(jar_path), *(str(info.libraries[lib_spec].absolute()) for lib_spec in processor.class_path)]),
                main_class,
                *(replace_install_args(arg) for arg in processor.args)
            ]

            watcher.handle(ForgePostProcessingEvent(task))

            completed = subprocess.run(args, cwd=self.context.work_dir, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            if completed.returncode != 0:
                raise ValueError("ERROR", completed.stdout)
            
            # If there are sha1, check them.
            for lib_name, expected_sha1 in processor.sha1.items():
                lib_name = replace_install_args(lib_name)
                expected_sha1 = replace_install_args(expected_sha1)
                with open(lib_name, "rb") as fp:
                    actual_sha1 = calc_input_sha1(fp)
                    if actual_sha1 != expected_sha1:
                        raise ValueError(f"invalid sha1 for '{lib_name}', got {actual_sha1}, expected {expected_sha1}")
        
        # Finally, remove the temporary directory.
        shutil.rmtree(info.tmp_dir, ignore_errors=True)

        watcher.handle(ForgePostProcessedEvent())


class _NeoForgeVersion(ForgeVersion):

    def __init__(self, neoforge_version: str = "release", *,
        context: Optional[Context] = None,
        prefix: str = "neoforge",
    ) -> None:
        super().__init__(neoforge_version, context=context, prefix=prefix)
    
    def _resolve_version(self, watcher: Watcher) -> None:
        
        # Maybe "release" or "snapshot", we process this first.
        self.forge_version = self.manifest.filter_latest(self.forge_version)[0]

        # The forge version is not fully specified.
        if "-" not in self.forge_version and self.forge_version.startswith("1."):

            watcher.handle(ForgeResolveEvent(self.forge_version, True, _api = "neoforge"))
            full_version = _request_neoforge_version(self.forge_version)

            if full_version is None:
                raise VersionNotFoundError(f"{self.prefix}-{self.forge_version}-???")
            
            self.forge_version = full_version
            watcher.handle(ForgeResolveEvent(self.forge_version, False, _api = "neoforge"))
        
        # Finally define the full version id.
        self.version = f"{self.prefix}-{self.forge_version}"
        
        # This is using the legacy forge artifact for 1.20.1 only.
        forge_artifact = _FORGE_ARTIFACT if self.forge_version.startswith("1.20.1-") else _NEO_FORGE_ARTIFACT

        self._forge_repo_url = _NEO_FORGE_REPO_URL
        self._forge_installer_spec = LibrarySpecifier(_NEO_FORGE_GROUP, forge_artifact, self.forge_version, classifier="installer")


class _ForgePostProcessor:
    """Describe the execution model of a post process.
    """
    __slots__ = "spec", "class_path", "args", "sha1"
    def __init__(self, spec: LibrarySpecifier, class_path: List[LibrarySpecifier], args: List[str], sha1: Dict[str, str]) -> None:
        self.spec = spec
        self.class_path = class_path
        self.args = args
        self.sha1 = sha1


class _ForgePostInfo:
    """Internal state, used only when forge installer is "modern" (>= 1.12.2-14.23.5.2851)
    describing data and post processors.
    """

    def __init__(self, tmp_dir: Path) -> None:
        self.tmp_dir = tmp_dir
        self.variables: Dict[str, str] = {}   # Data for variable replacements.
        self.libraries: Dict[LibrarySpecifier, Path] = {}  # Install-time libraries.
        self.processors: List[_ForgePostProcessor] = []


class ForgeInstallError(Exception):
    """Errors that can happen while trying to install forge.
    """

    INSTALL_PROFILE_NOT_FOUND = "install_profile_not_found"
    VERSION_METADATA_NOT_FOUND = "version_meta_not_found"

    def __init__(self, version: str, code: str):
        self.version = version
        self.code = code
    
    def __str__(self) -> str:
        return repr((self.version, self.code))


class ForgeResolveEvent:
    """Event triggered when the full forge version has been/is being resolved. 
    The 'alias' attribute specifies if an alias version is being resolved, if false the
    resolving has finished and we'll try to install the given version.
    """
    __slots__ = "forge_version", "alias", "_api"

    def __init__(self, forge_version: str, alias: bool, *, _api = "forge") -> None:
        self.forge_version = forge_version
        self.alias = alias
        self._api = _api

class ForgePostProcessingEvent:
    """Event triggered when a post processing task is starting.
    """
    __slots__ = "task",
    def __init__(self, task: str) -> None:
        self.task = task

class ForgePostProcessedEvent:
    """Event triggered when forge post processing has finished, the game is ready to run.
    """
    __slots__ = tuple()


def request_promo_versions() -> Dict[str, str]:
    """Request recommended and latest versions for each supported game release, this only
    works for forge version.
    """
    return http_request("GET", "https://files.minecraftforge.net/net/minecraftforge/forge/promotions_slim.json", 
        accept="application/json").json()["promos"]


def _request_neoforge_version(game_version: str) -> Optional[str]:
    """Request a neoforge version from a game version. Because of the NeoForge API 
    returning any matching version number, this function will also check that the returned
    version is of the right game version.
    """

    game_version_parts = game_version.split(".")
    if len(game_version_parts) < 2 or len(game_version_parts) > 3:
        return None
    
    # If the "super-major" version is not "1", abort...
    if game_version_parts[0] != "1":
        return None
    
    # Special case for the first version NeoForged was introduced.
    if game_version_parts == ["1", "20", "1"]:
        url = "https://maven.neoforged.net/api/maven/latest/version/releases/net%2Fneoforged%2Fforge?filter=1.20.1-"
    else:
        # Just keep major and minor version number and construct the neoforge version prefix.
        filter_version = ".".join(game_version_parts[1:3])
        url = f"https://maven.neoforged.net/api/maven/latest/version/releases/net%2Fneoforged%2Fneoforge?filter={url_parse.quote(filter_version)}"

    try:
        ret = http_request("GET", url, accept="application/json").json()
    except HttpError as err:
        if err.res.status != 404:
            raise
        return None

    return ret.get("version")


def request_maven_versions() -> List[str]:
    """Internal function that parses maven metadata of forge in order to get all 
    supported forge versions.
    """

    text = http_request("GET", f"{_FORGE_REPO_URL}/net/minecraftforge/forge/maven-metadata.xml", 
        accept="application/xml").text()
    
    versions = list()
    last_idx = 0

    # It's not really correct to parse XML like this, but I find this
    # acceptable since the schema is well known and it should be a
    # little bit easier to do thing like this.
    while True:
        start_idx = text.find("<version>", last_idx)
        if start_idx == -1:
            break
        end_idx = text.find("</version>", start_idx + 9)
        if end_idx == -1:
            break
        versions.append(text[(start_idx + 9):end_idx])
        last_idx = end_idx + 10

    return versions


def request_install_jar(version: str) -> ZipFile:
    """deprecated"""
    res = http_request("GET", f"{_FORGE_REPO_URL}/net/minecraftforge/forge/{version}/forge-{version}-installer.jar",
        accept="application/java-archive")
    
    return ZipFile(BytesIO(res.data))


def zip_extract_file(zf: ZipFile, entry_path: str, dst_path: Path):
    """Special function used to extract a specific file entry to a destination. 
    This is different from ZipFile.extract because the latter keep the full entry's path.
    """
    dst_path.parent.mkdir(parents=True, exist_ok=True)
    with zf.open(entry_path) as src, dst_path.open("wb") as dst:
        shutil.copyfileobj(src, dst)

#
# Following classes are deprecated public class that should never have been private...
#

class ForgePostProcessor:
    """deprecated: This class should have been private...
    """
    __slots__ = "jar_name", "class_path", "args", "sha1"
    def __init__(self, jar_name: str, class_path: List[str], args: List[str], sha1: Dict[str, str]) -> None:
        self.jar_name = jar_name
        self.class_path = class_path
        self.args = args
        self.sha1 = sha1

class ForgePostInfo:
    """deprecated: This class should have been private..
    """

    def __init__(self, tmp_dir: Path) -> None:
        self.tmp_dir = tmp_dir
        self.variables: Dict[str, str] = {}   # Data for variable replacements.
        self.libraries: Dict[str, Path] = {}  # Install-time libraries.
        self.processors: List[ForgePostProcessor] = []
